# @Author: Thomas Firmin <ThomasFirmin>
# @Date:   2022-05-03T15:41:48+02:00
# @Email:  thomas.firmin@univ-lille.fr
# @Project: Zellij
# @Last modified by:   ThomasFirmin
# @Last modified time: 2022-05-03T15:44:53+02:00
# @License: CeCILL-C (http://www.cecill.info/index.fr.html)
# @Copyright: Copyright (C) 2022 Thomas Firmin


import numpy as np
import abc
import copy


class Tree_search(object):

    """Tree_search

    Tree_search is an abstract class which determines how to explore the fractal rooted tree, builded during Fractal Decomposition.
    It is based on the OPEN/CLOSED lists algorithm.

    Attributes
    ----------

    open : list[Fractal]
        Open list containing not explored nodes from the fractal rooted tree.

    close : list[Fractal]
        Close list containing explored nodes from the fractal rooted tree.

    max_depth : int
        Maximum depth of the fractal rooted tree.

    Methods
    -------
    add(c)
        Add a node c to the fractal tree

    get_next()
        Get the next node to evaluate

    See Also
    --------
    Fractal : Abstract class defining what a fractal is.
    FDA : Fractal Decomposition Algorithm
    """

    def __init__(self, open, max_depth):

        """__init__(self,open,max_depth)

        Parameters
        ----------
        open : list[Fractal]
            Initial Open list containing not explored nodes from the fractal rooted tree.

        max_depth : int
            maximum depth of the fractal rooted tree.

        """

        ##############
        # PARAMETERS #
        ##############

        self.open = open
        self.close = []
        self.max_depth = max_depth

    @abc.abstractmethod
    def add(self, c):
        """__init__(open,max_depth)

        Parameters
        ----------
        c : Fractal
            Add a new fractal to the tree

        """
        pass

    @abc.abstractmethod
    def get_next(self):
        """__init__(open, max_depth)

        Returns
        -------

        continue : boolean
            If True determine if the open list has been fully explored or not

        nodes : {list[Fractal], -1}
            if -1 no more nodes to explore, else return a list of the next node to explore
        """
        pass


class Breadth_first_search(Tree_search):
    """Breadth_first_search

    Breadth First Search algorithm (BFS), computationally inefficient with fractal decomposition algorithm, because it is a greedy algorithm doing only exploration of the fractal tree,\
     exploring entirely each fractal level.

    Attributes
    ----------

    open : list[Fractal]
        Initial Open list containing not explored nodes from the fractal rooted tree.

    max_depth : int
        maximum depth of the fractal rooted tree.

    Q : int, default=1
        Q-Breadth_first_search, at each get_next, tries to return Q nodes.

    reverse : boolean, default=False
        if False do a descending sort the open list, else do an ascending sort

    Methods
    -------

    add(c)
        Add a node c to the fractal tree

    get_next()
        Get the next node to evaluate

    See Also
    --------
    Fractal : Abstract class defining what a fractal is.
    FDA : Fractal Decomposition Algorithm
    Tree_search : Base class
    Depth_first_search : Tree search Depth based startegy
    """

    def __init__(self, open, max_depth, Q=1, reverse=False):

        """__init__(open, max_depth, Q=1, reverse=False)

        Parameters
        ----------
        open : list[Fractal]
            Initial Open list containing not explored nodes from the fractal rooted tree.

        max_depth : int
            maximum depth of the fractal rooted tree.

        Q : int, default=1
            Q-Breadth_first_search, at each get_next, tries to return Q nodes.

        reverse : boolean, default=False
            if False do a descending sort the open list, else do an ascending sort

        """

        super().__init__(open, max_depth)

        ##############
        # PARAMETERS #
        ##############

        self.Q = Q
        self.reverse = reverse

        #############
        # VARIABLES #
        #############

        self.next_frontier = []

    def add(self, c):
        self.next_frontier.append(c)

    def get_next(self):

        if len(self.next_frontier) > 0:
            self.open = sorted(
                self.next_frontier + self.open,
                reverse=self.reverse,
                key=lambda x: x.level,
            )[:]
            self.next_frontier = []

        if len(self.open) > 0:

            idx_min = np.min([len(self.open), self.Q])

            self.close += self.open[0:idx_min]

            for _ in range(idx_min):
                self.open.pop(0)

            return True, self.close[-idx_min:]

        else:
            return False, -1


class Depth_first_search(Tree_search):
    """Depth_first_search

    Depth First Search algorithm (DFS), computationally inefficient with fractal decomposition algorithm, because it is a greedy algorithm doing only exploitation of the fractal tree,\
    by trying to go has deep as possible.

    Attributes
    ----------

    open : list[Fractal]
        Initial Open list containing not explored nodes from the fractal rooted tree.

    max_depth : int
        maximum depth of the fractal rooted tree.

    Q : int, default=1
        Q-Depth_first_search, at each get_next, tries to return Q nodes.

    reverse : boolean, default=False
        if False do a descending sort the open list, else do an ascending sort

    Methods
    -------
    add(self,c)
        Add a node c to the fractal tree

    get_next(self)
        Get the next node to evaluate

    See Also
    --------
    Fractal : Abstract class defining what a fractal is.
    FDA : Fractal Decomposition Algorithm
    Tree_search : Base class
    Breadth_first_search : Tree search Breadth based startegy
    Cyclic_best_first_search : Hybrid between DFS and BestFS
    """

    def __init__(self, open, max_depth, Q=1, reverse=False):
        """__init__(open, max_depth, Q=1, reverse=False)

        Parameters
        ----------
        open : list[Fractal]
            Initial Open list containing not explored nodes from the fractal rooted tree.

        max_depth : int
            maximum depth of the fractal rooted tree.

        Q : int, default=1
            Q-Depth_first_search, at each get_next, tries to return Q nodes.

        reverse : boolean, default=False
            if False do a descending sort the open list, else do an ascending sort

        """
        super().__init__(open, max_depth)

        ##############
        # PARAMETERS #
        ##############

        self.reverse = reverse
        self.Q = Q

        #############
        # VARIABLES #
        #############

        self.next_frontier = []

    def add(self, c):

        self.next_frontier.append(c)

    def get_next(self):

        if len(self.next_frontier) > 0:
            self.open = (
                sorted(
                    self.next_frontier,
                    reverse=self.reverse,
                    key=lambda x: x.score,
                )[:]
                + self.open
            )
            self.next_frontier = []

        if len(self.open) > 0:

            idx_min = np.min([len(self.open), self.Q])

            self.close += self.open[0:idx_min]

            for _ in range(idx_min):
                self.open.pop(0)

            return True, self.close[-idx_min:]

        else:
            return False, -1


class Best_first_search(Tree_search):

    """Best_first_search

    Best First Search algorithm (BestFS), BestFS is better than BFS and DFS, because it tries to explore and exploit only the best current node in the tree.

    Attributes
    ----------

    open : list[Fractal]
        Initial Open list containing not explored nodes from the fractal rooted tree.

    max_depth : int
        maximum depth of the fractal rooted tree.

    Q : int, default=1
        Q-Best_first_search, at each get_next, tries to return Q nodes.

    reverse : boolean, default=False
        if False do a descending sort the open list, else do an ascending sort

    Methods
    -------
    add(self,c)
        Add a node c to the fractal tree

    get_next(self)
        Get the next node to evaluate

    See Also
    --------
    Fractal : Abstract class defining what a fractal is.
    FDA : Fractal Decomposition Algorithm
    Tree_search : Base class
    Beam_search : Memory efficient tree search algorithm based on BestFS
    Cyclic_best_first_search : Hybrid between DFS and BestFS
    """

    def __init__(self, open, max_depth, Q=1, reverse=False):
        """__init__(self, open, max_depth, Q=1, reverse=False)

        Parameters
        ----------
        open : list[Fractal]
            Initial Open list containing not explored nodes from the fractal rooted tree.

        max_depth : int
            maximum depth of the fractal rooted tree.

        Q : int, default=1
            Q-Best_first_search, at each get_next, tries to return Q nodes.

        reverse : boolean, default=False
            if False do a descending sort the open list, else do an ascending sort

        """
        super().__init__(open, max_depth)

        ##############
        # PARAMETERS #
        ##############

        self.reverse = reverse
        self.Q = Q

        #############
        # VARIABLES #
        #############

        self.next_frontier = []

    def add(self, c):

        self.next_frontier.append(c)

    def get_next(self):

        if len(self.next_frontier) > 0:
            self.open = sorted(
                self.open
                + sorted(
                    self.next_frontier,
                    reverse=self.reverse,
                    key=lambda x: x.score,
                )[:],
                reverse=self.reverse,
                key=lambda x: x.score,
            )
            self.next_frontier = []

        if len(self.open) > 0:

            idx_min = np.min([len(self.open), self.Q])
            self.close += self.open[0:idx_min]

            for _ in range(idx_min):
                self.open.pop(0)

            return True, self.close[-idx_min:]

        else:
            return False, -1


class Beam_search(Tree_search):

    """Beam_search

    Beam Search algorithm (BS). BS is an improvement of BestFS. It includes a beam length (resp. open list length),\
    which allows to prune the worst nodes and only keep in memory a certain number of the best found nodes.

    Attributes
    ----------

    open : list[Fractal]
        Initial Open list containing not explored nodes from the fractal rooted tree.

    max_depth : int
        maximum depth of the fractal rooted tree.

    Q : int, default=1
        Q-Beam_search, at each get_next, tries to return Q nodes.

    reverse : boolean, default=False
        if False do a descending sort the open list, else do an ascending sort

    Methods
    -------
    add(c)
        Add a node c to the fractal tree

    get_next()
        Get the next node to evaluate

    beam_length : int, default=10
        Determines the length of the open list for memory and prunning issues.

    See Also
    --------
    Fractal : Abstract class defining what a fractal is.
    FDA : Fractal Decomposition Algorithm
    Tree_search : Base class
    Best_first_search : Tree search algorithm based on the best node from the open list
    Cyclic_best_first_search : Hybrid between DFS and BestFS, which can also perform pruning.
    """

    def __init__(self, open, max_depth, Q=1, reverse=False, beam_length=10):

        """__init__(open, max_depth, Q=1, reverse=False)

        Parameters
        ----------
        open : list[Fractal]
            Initial Open list containing not explored nodes from the fractal rooted tree.

        max_depth : int
            maximum depth of the fractal rooted tree.

        Q : int, default=1
            Q-Beam_search, at each get_next, tries to return Q nodes.

        reverse : boolean, default=False
            If False do a descending sort the open list, else do an ascending sort

        beam_length : int, default=10
            Determines the length of the open list for memory and prunning issues.

        """

        super().__init__(open, max_depth)

        ##############
        # PARAMETERS #
        ##############

        self.reverse = reverse
        self.open = open
        self.Q = Q
        self.beam_length = beam_length

        #############
        # VARIABLES #
        #############

        self.next_frontier = []

    def add(self, c):

        self.next_frontier.append(c)

    def get_next(self):

        if len(self.next_frontier) > 0:
            self.open = sorted(
                self.next_frontier
                + sorted(
                    self.open, reverse=self.reverse, key=lambda x: x.score
                ),
                reverse=self.reverse,
                key=lambda x: x.score,
            )[: self.beam_length]
            self.next_frontier = []

        if len(self.open) > 0:

            idx_min = np.min([len(self.open), self.Q])
            self.close += self.open[0:idx_min]

            for _ in range(idx_min):
                self.open.pop(0)

            return True, self.close[-idx_min:]

        else:
            return False, -1


class Diverse_best_first_search(Tree_search):

    """Diverse_best_first_search

    Diverse Best First Search (DBFS). DBFS is an improvement of BestFS. When a node is badly evaluated, this one has no more chance to be explored.\
    DBFS tries to tackle this problem by randomly selecting nodes according to a probability computed with its heuristic value (score) and its parents scores,\
    or according to a probability P.

    Attributes
    ----------

    open : list[Fractal]
        Initial Open list containing not explored nodes from the fractal rooted tree.

    max_depth : int
        maximum depth of the fractal rooted tree.

    Q : int, default=1
        Q-Diverse_best_first_search, at each get_next, tries to return Q nodes.

    reverse : boolean, default=False
        if False do a descending sort the open list, else do an ascending sort

    P : float, default=0.1
        Probability to select a random node from the open list. Determine how random the selection must be. The higher it is, the more exploration DBFS does.

    T : float, default=0.5
        Influences the probability of a node to be selected according to its score compared to the best score from the open list.

    Methods
    -------
    add(c)
        Add a node c to the fractal tree

    get_next()
        Get the next node to evaluate

    See Also
    --------
    Fractal : Abstract class defining what a fractal is.
    FDA : Fractal Decomposition Algorithm
    Tree_search : Base class
    Best_first_search : Tree search algorithm based on the best node from the open list
    Epsilon_greedy_search : Based on BestFS, allows to randomly select a node.
    """

    def __init__(self, open, max_depth, Q=1, reverse=False, P=0.1, T=0.5):

        """__init__(open, max_depth, Q=1, reverse=False)

        Parameters
        ----------
        open : list[Fractal]
            Initial Open list containing not explored nodes from the fractal rooted tree.

        max_depth : int
            maximum depth of the fractal rooted tree.

        Q : int, default=1
            Q-Diverse_best_first_search, at each get_next, tries to return Q nodes.

        reverse : boolean, default=False
            If False do a descending sort the open list, else do an ascending sort

        P : float, default=0.1
            Probability to select a random node from the open list. Determine how random the selection must be. The higher it is, the more exploration DBFS does.

        T : float, default=0.5
            Influences the probability of a node to be selected according to its score compared to the best score from the open list.
        """

        super().__init__(open, max_depth)

        ##############
        # PARAMETERS #
        ##############

        self.open = open
        self.reverse = reverse
        self.Q = Q
        self.P = P
        self.T = T

        #############
        # VARIABLES #
        #############
        self.next_frontier = []

        for i in self.open:
            i.g_value = i.min_score

    def add(self, c):

        c.g_value = c.min_score
        start = c.father

        while type(start.father) != str:
            c.g_value += start.father.min_score
            start = start.father

        self.next_frontier.append(c)

    def fetch_node(self):

        if len(self.open) > 1:
            p_total = 0

            h_values = [i.min_score for i in self.open]
            g_values = [i.g_value for i in self.open]

            p = []

            combination = []

            hmin, hmax = np.min(h_values), np.max(h_values)
            gmin, gmax = np.min(g_values), np.max(g_values)

            if np.random.random() < self.P:
                G = np.random.choice(g_values)
            else:
                G = gmax

            for h, g in zip(h_values, g_values):

                if g > G:
                    p.append(0)
                else:
                    p.append(self.T ** (h - hmin))

                p_total += p[-1]

            idx = np.random.choice(len(self.open), p=p / p_total)

            return idx

        else:
            return 0

    def get_next(self):

        if len(self.next_frontier) > 0:
            self.open = sorted(
                self.next_frontier
                + sorted(
                    self.open, reverse=self.reverse, key=lambda x: x.score
                ),
                reverse=self.reverse,
                key=lambda x: x.score,
            )
            self.next_frontier = []

        if len(self.open) > 0:

            idx = self.fetch_node()
            self.close += [self.open[idx]]
            self.open.pop(idx)

            return True, [self.close[-1]]

        else:
            return False, -1


class Cyclic_best_first_search(Tree_search):

    """Cyclic_best_first_search

    Cyclic Best First Search (CBFS). CBFS is an hybrid of DFS and BestFS. First, CBFS tries to reach a leaf of the fractal tree to quickly determine a base score.
    Then CBFS will do pruning according to this value, and it will decompose the problem into subproblems by inserting nodes into contours (collection of unexplored subproblems).
    At each iteration CBFS select the best subproblem according to an heuristic value. Then the child subproblems will be inserted into their respective contours, according to a labelling function.

    Attributes
    ----------

    open : list[Fractal]
        Initial Open list containing not explored nodes from the fractal rooted tree.

    max_depth : int
        maximum depth of the fractal rooted tree.

    Q : int, default=1
        Q-Cyclic_best_first_search, at each get_next, tries to return Q nodes.

    reverse : boolean, default=False
        if False do a descending sort the open list, else do an ascending sort

    Methods
    -------
    add(c)
        Add a node c to the fractal tree

    get_next()
        Get the next node to evaluate

    See Also
    --------
    Fractal : Abstract class defining what a fractal is.
    FDA : Fractal Decomposition Algorithm
    Tree_search : Base class
    Best_first_search : Tree search algorithm based on the best node from the open list
    Depth_first_search : Tree search Depth based startegy
    """

    def __init__(self, open, max_depth, Q=1, reverse=False):

        """__init__(open, max_depth, Q=1, reverse=False)

        Parameters
        ----------
        open : list[Fractal]
            Initial Open list containing not explored nodes from the fractal rooted tree.

        max_depth : int
            maximum depth of the fractal rooted tree.

        Q : int, default=1
            Q-Cyclic_best_first_search, at each get_next, tries to return Q nodes.

        reverse : boolean, default=False
            If False do a descending sort the open list, else do an ascending sort

        """

        super().__init__(open, max_depth)

        ##############
        # PARAMETERS #
        ##############

        self.reverse = reverse
        self.Q = Q

        #############
        # VARIABLES #
        #############

        self.next_frontier = []

        self.L = [False] * (self.max_depth + 1)
        self.L[0] = True
        self.i = 0
        self.contour = [[] for i in range(self.max_depth + 1)]
        self.contour[0] = open

        self.best_scores = float("inf")
        self.first_complete = False

    def add(self, c):

        # Verify if a node must be pruned or not.
        # A node can be pruned only if at least one exploitation has been made
        if not self.first_complete:
            self.next_frontier.append(c)

            if c.level == self.max_depth:
                self.first_complete = True
                self.best_score = c.min_score
        else:
            if c.min_score < self.best_score:
                self.best_score = c.min_score
                self.next_frontier.append(c)

    def get_next(self):

        if len(self.next_frontier) > 0:

            modified_levels = []
            for h in self.next_frontier:

                self.contour[h.level].append(h)
                modified_levels.append(h.level)

                if not self.L[h.level]:
                    self.L[h.level] = True

            modified_levels = np.unique(modified_levels)
            for l in modified_levels:

                self.contour[l] = sorted(
                    self.contour[l], reverse=self.reverse, key=lambda x: x.score
                )

            self.next_frontier = []

        if np.any(self.L):

            search = True
            found = True

            l = 0
            i = -1

            while l < len(self.L) and search:

                if self.L[l]:

                    if found:
                        i = l
                        found = False

                    if l > self.i:
                        self.i = l
                        search = False

                l += 1

            if search and not found:
                self.i = i

            idx_min = np.min([len(self.contour[self.i]), self.Q])

            self.close += self.contour[self.i][0:idx_min]

            for _ in range(idx_min):
                self.contour[self.i].pop(0)

            if len(self.contour[self.i]) == 0:
                self.L[self.i] = False

            return True, self.close[-idx_min:]

        else:
            return False, -1


class Epsilon_greedy_search(Tree_search):

    """Epsilon_greedy_search

    Epsilon Greedy Search (EGS). EGS is an improvement of BestFS. At each iteration nodes are selected randomly or according to their best score.

    Attributes
    ----------

    open : list[Fractal]
        Initial Open list containing not explored nodes from the fractal rooted tree.

    max_depth : int
        maximum depth of the fractal rooted tree.

    Q : int, default=1
        Q-Epsilon_greedy_search, at each get_next, tries to return Q nodes.

    reverse : boolean, default=False
        if False do a descending sort the open list, else do an ascending sort

    epsilon : float, default=0.1
        Probability to select a random node from the open list. Determine how random the selection must be. The higher it is, the more exploration EGS does.

    Methods
    -------
    add(c)
        Add a node c to the fractal tree

    get_next()
        Get the next node to evaluate

    See Also
    --------
    Fractal : Abstract class defining what a fractal is.
    FDA : Fractal Decomposition Algorithm
    Tree_search : Base class
    Best_first_search : Tree search algorithm based on the best node from the open list
    Diverse_best_first_search : Tree search strategy based on an adaptative probability to select random nodes.
    """

    def __init__(self, open, max_depth, reverse=False, epsilon=0.1):

        """__init__(open, max_depth, Q=1, reverse=False)

        Parameters
        ----------
        open : list[Fractal]
            Initial Open list containing not explored nodes from the fractal rooted tree.

        max_depth : int
            maximum depth of the fractal rooted tree.

        Q : int, default=1
            Q-Epsilon_greedy_search, at each get_next, tries to return Q nodes.

        reverse : boolean, default=False
            If False do a descending sort the open list, else do an ascending sort

        epsilon : float, default=0.1
            Probability to select a random node from the open list. Determine how random the selection must be. The higher it is, the more exploration EGS does.

        """

        super().__init__(open, max_depth)

        ##############
        # PARAMETERS #
        ##############

        self.reverse = reverse
        self.epsilon = epsilon

        #############
        # VARIABLES #
        #############

        self.next_frontier = []

    def add(self, c):

        self.next_frontier.append(c)

    def get_next(self):

        if len(self.next_frontier) > 0:
            self.open = sorted(
                self.open
                + sorted(
                    self.next_frontier,
                    reverse=self.reverse,
                    key=lambda x: x.score,
                )[:],
                reverse=self.reverse,
                key=lambda x: x.score,
            )
            self.next_frontier = []

        if len(self.open) > 0:

            if np.random.random() > self.epsilon:

                self.close += [self.open[0]]
                self.open.pop(0)

            else:

                if len(self.open) > 1:
                    idx = np.random.randint(1, len(self.open))
                    self.close += [self.open[idx]]
                    self.open.pop(idx)
                else:
                    self.close += [self.open[0]]
                    self.open.pop(0)

            return True, [self.close[-1]]

        else:
            return False, -1


class Potentially_Optimal_Rectangle(Tree_search):

    """Potentially_Optimal_Rectangle

    Potentially Optimal Rectangle algorithm (POR), is a the selection strategy comming from DIRECT.

    Attributes
    ----------

    open : list[Fractal]
        Initial Open list containing not explored nodes from the fractal rooted tree.

    max_depth : int
        maximum depth of the fractal rooted tree.

    Q : int, default=1
        Q-Best_first_search, at each get_next, tries to return Q nodes.

    reverse : boolean, default=False
        if False do a descending sort the open list, else do an ascending sort

    Methods
    -------
    add(self,c)
        Add a node c to the fractal tree

    get_next(self)
        Get the next node to evaluate

    See Also
    --------
    Fractal : Abstract class defining what a fractal is.
    FDA : Fractal Decomposition Algorithm
    Tree_search : Base class
    Beam_search : Memory efficient tree search algorithm based on BestFS
    Cyclic_best_first_search : Hybrid between DFS and BestFS
    """

    def __init__(self, open, max_depth, error=1e-4):
        """__init__(self, open, max_depth, Q=1, reverse=False, error=1e-4)

        Parameters
        ----------
        open : list[Fractal]
            Initial Open list containing not explored nodes from the fractal rooted tree.

        max_depth : int
            maximum depth of the fractal rooted tree.

        Q : int, default=1
            Q-Best_first_search, at each get_next, tries to return Q nodes.

        reverse : boolean, default=False
            if False do a descending sort the open list, else do an ascending sort

        error : float, default=1e-4
            Small value which determines when an evaluation should be considered as good as the best solution found so far.

        """
        super().__init__(open, max_depth)

        ##############
        # PARAMETERS #
        ##############

        self.error = error

        #############
        # VARIABLES #
        #############

        self.next_frontier = []

        self.best_score = float("inf")

        self.build = np.vectorize(
            lambda x: (x.min_score, np.linalg.norm(x.center - x.lo_bounds))
        )

    def add(self, c):

        self.next_frontier.append(c)

        if c.min_score < self.best_score:
            self.best_score = c.min_score

    def get_next(self):

        if len(self.next_frontier) > 0:
            idx = self.optimal()

            for i, j in enumerate(idx):
                self.open.append(self.next_frontier.pop(j - i))

        if len(self.open) > 0:

            idx_min = len(self.open)

            for _ in range(idx_min):
                self.close.append(self.open.pop(0))

            return True, self.close[-idx_min:]

        else:
            return False, -1

    def optimal(self):

        res = []

        a, d = self.build(self.next_frontier)

        p1 = a.reshape((len(a), 1)) - a
        p2 = d.reshape((len(d), 1)) - d
        K = p1 / p2

        K2 = (a - self.best_score + self.error * np.abs(self.best_score)) / d

        for i in range(len(a)):
            U = K[i][p2[i] < 0]
            if len(U) == 0:
                res.append(i)
            else:
                minU = np.nanmin(U)
                if minU > 0:
                    L = K[i][p2[i] >= 0]
                    maxL = np.maximum(np.nanmax(L), K2[i])
                    if maxL > 0 and maxL <= minU:
                        res.append(i)

        return res



# @Author: Thomas Firmin <ThomasFirmin>
# @Date:   2022-05-03T15:41:48+02:00
# @Email:  thomas.firmin@univ-lille.fr
# @Project: Zellij
# @Last modified by:   ThomasFirmin
# @Last modified time: 2022-05-03T15:45:43+02:00
# @License: CeCILL-C (http://www.cecill.info/index.fr.html)
# @Copyright: Copyright (C) 2022 Thomas Firmin


import numpy as np
from zellij.core.metaheuristic import Metaheuristic
from zellij.core.fractals import Hypersphere

import logging

logger = logging.getLogger("zellij.ILS")

# Intensive local search
class ILS(Metaheuristic):

    """ILS

    Intensive local search is an exploitation algorithm comming from original FDA paper.
    It evaluate a point in each dimension arround an initial solution.
    Distance of the computed point to the initial one is decreasing according to a reduction rate.
    At each iteration the algorithm moves to the best solution found.

    Attributes
    ----------

    up_bounds : list
        List of float containing the upper bounds of the search space converted to continuous.
    lo_bounds : list
        List of float containing the lower bounds of the search space converted to continuous.
    center : float
        List of floats containing the coordinates of the search space center converted to continuous.
    radius : float
        List of floats containing the radius for each dimensions of the search space converted to continuous.
    red_rate : float
        determine the step reduction rate ate each iteration.
    precision : float
        dtermine the stopping criterion. When the step is lower than <precision> the algorithm stops.

    Methods
    -------

    run(self, n_process=1)
        Runs ILS


    See Also
    --------
    Metaheuristic : Parent class defining what a Metaheuristic is.
    LossFunc : Describes what a loss function is in Zellij
    Searchspace : Describes what a loss function is in Zellij
    """

    def __init__(
        self,
        loss_func,
        search_space,
        f_calls,
        red_rate=0.5,
        precision=1e-5,
        verbose=True,
    ):

        """__init__(loss_func, search_space, f_calls,save=False,verbose=True)

        Initialize ILS class

        Parameters
        ----------
        loss_func : Loss
            Loss function to optimize. must be of type f(x)=y

        search_space : Searchspace
            Search space object containing bounds of the search space.

        f_calls : int
            Maximum number of loss_func calls

        red_rate : float, default=0.5
            determine the step reduction rate ate each iteration.

        precision : float, default=1e-5
            dtermine the stopping criterion. When the step is lower than <precision> the algorithm stops.

        verbose : boolean, default=True
            Algorithm verbosity

        """

        super().__init__(loss_func, search_space, f_calls, verbose)

        self.red_rate = red_rate
        self.precision = precision

    def run(self, X0=None, Y0=None, H=None, n_process=1):

        """run(X0=None, Y0=None, H=None, n_process=1)

        Parameters
        ----------
        X0 : list[float], optional
            Initial solution. If None, a Fractal must be given (H!=None)
        Y0 : {int, float}, optional
            Score of the initial solution
            Determine the starting point of the chaotic map.
        H : Fractal, optional
            When used by FDA, a fractal corresponding to the current subspace is given
        n_process : int, default=1
            Determine the number of best solution found to return.

        Returns
        -------
        best_sol : list[float]
            Returns a list of the <n_process> best found points to the continuous format

        best_scores : list[float]
            Returns a list of the <n_process> best found scores associated to best_sol

        """

        if H:
            assert isinstance(H, Hypersphere), logger.error(
                f"ILS should use Hyperspheres, got {H.__class__.__name__}"
            )

        # logging
        logger.info("Starting")

        self.build_bar(self.f_calls)

        scores = np.zeros(3, dtype=float)

        if X0:
            points = np.tile(X0, (3, 1))
            points[points > 1] = 1
            points[points < 0] = 0

        elif H:
            points = np.tile(H.center, (3, 1))
            points[points > 1] = 1
            points[points < 0] = 0
        else:
            raise ValueError("No starting point given to Simulated Annealing")

        if Y0:
            scores[0] = Y0
        else:
            scores[0] = self.loss_func(
                self.search_space.convert_to_continuous([points[0]], True, True)
            )[0]

        step = H.radius

        while step > self.precision and self.loss_func.calls < self.f_calls:
            i = 0
            improvement = False
            # logging
            logger.debug(f"ILS {step}>{self.precision}")

            while (
                i < self.search_space.n_variables
                and self.loss_func.calls < self.f_calls
            ):

                # logging
                logger.debug(f"Evaluating dimension {i}")

                self.pending_pb(2)

                walk = points[0][i] + step
                points[1][i] = walk
                points[1][points[1] > 1] = 1
                points[1][points[1] < 0] = 0

                walk = points[0][i] - step
                points[2][i] = walk
                points[2][points[2] > 1] = 1
                points[2][points[2] < 0] = 0

                scores[1:] = self.loss_func(
                    self.search_space.convert_to_continuous(
                        points[1:], True, True
                    )
                )

                min_index = np.argmin(scores)

                if min_index != 0:
                    points = np.tile(points[min_index], (3, 1))
                    scores[0] = scores[min_index]
                    improvement = True

                self.update_main_pb(
                    2, explor=False, best=self.loss_func.new_best
                )
                self.meta_pb.update(2)

                i += 1

            if not improvement:
                step = self.red_rate * step

        # logging
        logger.info("Ending")
        self.close_bar()

        return points[0], scores[0]

    def show(self, filepath="", save=False):

        """show(filename="")

        Plots solutions and scores evaluated during the optimization

        Parameters
        ----------
        filename : str, default=None
            If a filepath is given, the method will read the file and will try to plot contents.

        save : boolean, default=False
            Save figures
        """

        super().show(filepath, save)



# @Author: Thomas Firmin <ThomasFirmin>
# @Date:   2022-05-03T15:41:48+02:00
# @Email:  thomas.firmin@univ-lille.fr
# @Project: Zellij
# @Last modified by:   ThomasFirmin
# @Last modified time: 2022-05-03T15:45:03+02:00
# @License: CeCILL-C (http://www.cecill.info/index.fr.html)
# @Copyright: Copyright (C) 2022 Thomas Firmin


from zellij.core.metaheuristic import Metaheuristic
from zellij.strategies.utils.chaos_map import Chaos_map
from zellij.strategies.utils.chaos_map import chaos_map_name
import zellij.utils.progress_bar as pb

import numpy as np

import logging

logger = logging.getLogger("zellij.CO")


class CGS(Metaheuristic):

    """Chaotic Global search

    CGS is an exploration :ref:`meta` using chaos to violently move in the :ref:`sp`.
    It is continuous optimization, so the :ref:`sp` is converted to continuous.
    To do so, it uses a :ref:`cmap`, such as Henon or Kent map.

    Attributes
    ----------
    level : int
        Chaotic level corresponds to the number of iteration of the chaotic map
    map : Chaos_map
        Chaotic map used to sample points. See Chaos_map object.
    up_bounds : list
        List of float containing the upper bounds of the search space converted to continuous
    lo_bounds : list
        List of float containing the lower bounds of the search space converted to continuous
    center : float
        List of floats containing the coordinates of the search space center converted to continuous
    radius : float
        List of floats containing the radius for each dimensions of the search space converted to continuous

    See Also
    --------
    :ref:`meta` : Parent class defining what a Metaheuristic is
    Chaotic_optimization : CGS is used here to perform an exploration
    CLS : Chaotic Local Search
    CFS : Chaotic Fine Search

    Examples
    --------
    >>> from zellij.core.loss_func import Loss
    >>> from zellij.core.search_space import Searchspace
    >>> from zellij.strategies.chaos_algorithm import CGS
    >>> from zellij.strategies.utils.chaos_map import Henon
    >>> from zellij.utils.benchmark import himmelblau
    ...
    >>> labels = ["a","b","c"]
    >>> types = ["R","R","R"]
    >>> values = [[-5, 5],[-5, 5],[-5, 5]]
    >>> sp = Searchspace(labels,types,values)
    >>> lf = Loss()(himmelblau)
    ...
    ...          # Henon(map size, dimensions)
    >>> chaosmap = Henon(250,3)
    ...                  # 4 points/iterations: 4x250=1000
    >>> cgs = CGS(lf, sp, 1000, 250, chaosmap)
    >>> cgs.run()
    >>> cgs.show()


    """

    def __init__(
        self, loss_func, search_space, f_calls, level, map, verbose=True
    ):
        """__init__(loss_func, search_space, f_calls, level, map, verbose=True)

        Initialize CGS class

        Parameters
        ----------
        loss_func : Loss
            Loss function to optimize. must be of type f(x)=y
        search_space : Searchspace
            Search space object containing bounds of the search space
        f_calls : int
            Maximum number of loss_func calls
        level : int
            Chaotic level corresponds to the number of iteration of the chaotic map
        map : Chaos_map
            Chaotic map used to sample points. See `Chaos_map` object.
        verbose : boolean, default=True
            Algorithm verbosity

        """

        ##############
        # PARAMETERS #
        ##############

        super().__init__(loss_func, search_space, f_calls, verbose)

        self.map = map.map
        self.level = level

        #############
        # VARIABLES #
        #############

        self.up_bounds = np.array(
            [1 for _ in range(self.search_space.n_variables)]
        )
        self.lo_bounds = np.array(
            [0 for _ in range(self.search_space.n_variables)]
        )

        # Working attributes, saved to avoid useless computations.
        self.up_plus_lo = self.up_bounds + self.lo_bounds
        self.up_m_lo = self.up_bounds - self.lo_bounds

        self.center = np.multiply(0.5, self.up_plus_lo)
        self.radius = np.multiply(0.5, self.up_m_lo)
        self.center_m_lo_bounds = self.center - self.lo_bounds

    def run(self, shift=1, H=None, n_process=1):

        """run(shift=1, H=None, n_process=1)

        Parameters
        ----------
        shift : int, default=1
            Determines the starting point of the chaotic map.

        H : Fractal, default=None
            When used by FDA, a fractal corresponding to the current subspace is given

        n_process : int, default=1
            Determines the number of best solution found to return.

        Returns
        -------
        best_sol : list[float]
            Returns a list of the <n_process> best found points to the continuous format

        best_scores : list[float]
            Returns a list of the <n_process> best found scores associated to best_sol

        """

        logger.info("CGS starting")

        self.build_bar(self.level)

        self.k = shift

        # For each level of chaos
        shift_map = (self.k - 1) * self.level
        points = np.empty((0, self.search_space.n_variables), dtype=float)

        n_points = self.loss_func.calls
        l = 0

        logger.info("CGS computing chaotic points")

        while l < self.level and n_points < self.f_calls:

            # Randomly select a parameter index of a solution
            d = np.random.randint(self.search_space.n_variables)

            # Apply 3 transformations on the selected chaotic variables
            r_mul_y = np.multiply(self.up_m_lo, self.map[l + shift_map])

            # xx = [np.add(self.center,r_mul_y), np.add(self.center,np.multiply(self.radius,np.multiply(2,y)-1)), np.subtract(self.up_bounds,r_mul_y)]

            # for each transformation of the chaotic variable
            # for x in xx:
            #
            #     x_ = np.subtract(self.up_plus_lo,x)
            #     sym = np.matrix([x,x,x_,x_])
            #     sym[1,d] = x_[d]
            #     sym[3,d] = x[d]
            #     points = np.append(points,sym,axis=0)
            #     n_points += 4

            xx = [self.lo_bounds + r_mul_y, self.up_bounds - r_mul_y]

            # for each transformation of the chaotic variable
            sym = np.array([xx[0], xx[1], xx[0], xx[1]])
            sym[2, d] = xx[1][d]
            sym[3, d] = xx[0][d]

            points = np.append(points, sym, axis=0)
            n_points += 4

            l += 1
            self.meta_pb.update()

        # Update progress bar
        self.pending_pb(len(points))

        logger.info("CGS evaluating chaotic points")
        ys = self.loss_func(
            self.search_space.convert_to_continuous(points, True),
            algorithm="CGS",
        )

        # Update progress bar
        self.update_main_pb(
            len(points), explor=True, best=self.loss_func.new_best
        )

        ys = np.array(ys)
        idx = np.array(np.argsort(ys))[:n_process]

        # best solution found
        best_sol = points[idx]
        best_scores = ys[idx]

        self.close_bar()

        logger.info("CGS ending")

        return best_sol, best_scores


class CLS(Metaheuristic):

    """Chaotic Local Search

    CLS is an exploitation :ref:`meta` using chaos to wiggle points arround an initial solution.\
     It uses a rotating polygon to distribute those points, a progressive and mooving zoom on the best solution found, to refine it.
    It is continuous optimization, so the :ref:`sp` is converted to continuous.
    To do so, it uses a :ref:`cmap`, such as Henon or Kent map.

    Attributes
    ----------
    level : int
        Chaotic level: the number of iteration of the chaotic map
    map : Chaos_map
        Chaotic map used to sample points. See Chaos_map object.
    polygon : int
        Vertex number of the rotating polygon (has an influence on the number of evaluated points)
    red_rate : float
        Reduction rate of the progressive zoom on the best solution found
    up_bounds : list
        List of float containing the upper bounds of the search space converted to continuous
    lo_bounds : list
        List of float containing the lower bounds of the search space converted to continuous
    center : float
        List of floats containing the coordinates of the search space center converted to continuous
    radius : float
        List of floats containing the radius for each dimensions of the search space converted to continuous

    See Also
    --------
    :ref:`meta` : Parent class defining what a Metaheuristic is
    Chaotic_optimization : CLS is used here to perform an exploitation
    CGS : Chaotic Global Search
    CFS : Chaotic Fine Search

    Examples
    --------
    >>> from zellij.core.loss_func import Loss
    >>> from zellij.core.search_space import Searchspace
    >>> from zellij.strategies.chaos_algorithm import CLS
    >>> from zellij.strategies.utils.chaos_map import Henon
    >>> from zellij.utils.benchmark import himmelblau
    ...
    >>> labels = ["a","b","c"]
    >>> types = ["R","R","R"]
    >>> values = [[-5, 5],[-5, 5],[-5, 5]]
    >>> sp = Searchspace(labels,types,values)
    >>> lf = Loss()(himmelblau)
    ...
    ...          # Henon(map size, dimensions)
    >>> chaosmap = Henon(50,3)
    ...                  # 2xpolygon points/iterations: 2x10x50=1000
    >>> cls = CLS(lf, sp, 1000, 50, 10, chaosmap)
    >>> point = sp.random_point()[0]
    >>> cls.run(point, lf([point])[0])
    >>> cls.show()

    """

    def __init__(
        self,
        loss_func,
        search_space,
        f_calls,
        level,
        polygon,
        map,
        verbose=True,
    ):

        """__init__(loss_func, search_space, f_calls, level, polygon, map, red_rate=0.5, verbose=True)

        Initialize CLS class

        Parameters
        ----------
        loss_func : Loss
            Loss function to optimize. must be of type f(x)=y
        search_space : Searchspace
            Search space object containing bounds of the search space
        f_calls : int
            Maximum number of loss_func calls
        level : int
            Chaotic level corresponds to the number of iteration of the chaotic map
        polygon : int
            Vertex number of the rotating polygon (has an influence on the number of evaluated points)
        map : Chaos_map
            Chaotic map used to sample points. See Chaos_map object.
        verbose : boolean, default=True
            Algorithm verbosity

        """

        ##############
        # PARAMETERS #
        ##############

        super().__init__(loss_func, search_space, f_calls, verbose)

        self.level = level
        self.polygon = polygon
        self.map = map.map
        self.red_rate = np.random.random()

        #############
        # VARIABLES #
        #############

        self.up_bounds = np.array([1 for _ in self.search_space.values])
        self.lo_bounds = np.array([0 for _ in self.search_space.values])

        self.up_plus_lo = self.up_bounds + self.lo_bounds
        self.up_m_lo = self.up_bounds - self.lo_bounds

        self.center = np.multiply(0.5, self.up_plus_lo)
        self.radius = np.multiply(0.5, self.up_m_lo)
        self.center_m_lo_bounds = self.center - self.lo_bounds

        trigo_val = 2 * np.pi / self.polygon
        self.H = [np.zeros(self.polygon), np.zeros(self.polygon)]

        for i in range(1, self.polygon + 1):
            # Initialize trigonometric part of symetric variables (CLS & CFS)
            self.H[0][i - 1] = np.cos(trigo_val * i)
            self.H[1][i - 1] = np.sin(trigo_val * i)

    def run(
        self, X0=None, Y0=None, chaos_level=0, shift=1, H=None, n_process=1
    ):

        """run(X0=None, Y0=None, chaos_level=0, shift=1, H=None, n_process=1)

        Parameters
        ----------
        X0 : list[float], optional
            Initial solution. If None, a Fractal must be given (H!=None)
        Y0 : {int, float}, optional
            Score of the initial solution
        chaos_level : int, default=0
            Determines at which level of the chaos map, the algorithm starts
        shift : int, default=1
            Determines the starting point of the chaotic map.
        H : Fractal, optional
            When used by FDA, a fractal corresponding to the current subspace is given
        n_process : int, default=1
            Determines the number of best solution found to return.

        Returns
        -------
        best_sol : list[float]
            Returns a list of the <n_process> best found points to the continuous format

        best_scores : list[float]
            Returns a list of the <n_process> best found scores associated to best_sol

        """

        logger.info("CLS starting")

        self.build_bar(self.level)

        if X0:
            self.X0 = np.array(
                self.search_space.convert_to_continuous([X0])[0], dtype=float
            )
        elif H:
            self.X0 = H.center
        else:
            raise ValueError("No starting point given to Simulated Annealing")

        if Y0:
            self.Y0 = Y0
        else:
            logger.info("CLS evaluating initial solution")
            self.Y0 = self.loss_func(
                self.search_space.convert_to_continuous([self.X0], True)
            )[0]

        self.k = shift
        self.chaos_level = chaos_level

        # Initialization
        shift = self.chaos_level * (self.k - 1) * self.level
        # Limits of the search space, if parameter greater than center, then = 1 else = -1, used to avoid overflow
        db = np.minimum(self.up_bounds - self.X0, self.X0 - self.lo_bounds)

        center_m_solution = self.center - self.X0
        points = np.empty((0, self.search_space.n_variables), dtype=float)

        n_points = self.loss_func.calls
        l = 0

        logger.info("CLS computing chaotic points")
        # for each level of chaos
        while l < self.level and n_points < self.f_calls:

            self.red_rate = np.random.random()

            # Local search area radius
            Rl = self.radius * self.red_rate

            # Decomposition vector
            d = np.random.randint(self.search_space.n_variables)

            # zoom speed
            gamma = 10 ** (-2 * self.red_rate * l) / (l + 1)

            # for each parameter of a solution, determine the improved radius
            xx = np.minimum(gamma * Rl, db)

            # Compute both chaotic variable of the polygonal model thanks to a chaotic map
            xv = [
                np.multiply(self.map[shift + l], xx),
                np.multiply(1 - self.map[shift + l], xx),
            ]

            # For both chaotic variable
            for x in xv:
                xi = np.outer(self.H[1], x)
                xi[:, d] = x[d] * self.H[0]
                xt = self.X0 + xi

                points = np.append(points, xt, axis=0)
                n_points += self.polygon

            l += 1
            self.meta_pb.update()

        # Update progress bar
        self.pending_pb(len(points))

        logger.info("CLS evaluating chaotic points")
        print(points)
        ys = self.loss_func(
            self.search_space.convert_to_continuous(points, True),
            algorithm="CLS",
        )

        # Update progress bar
        self.update_main_pb(
            len(points), explor=True, best=self.loss_func.new_best
        )

        ys = np.array(ys)
        idx = np.array(np.argsort(ys))[:n_process]

        # best solution found
        best_sol = points[idx]
        best_scores = ys[idx]

        self.close_bar()

        logger.info("CLS ending")

        return best_sol, best_scores


class CFS(Metaheuristic):

    """Chaotic Fine Search

    CFS is an exploitation :ref:`meta` using chaos to wiggle points arround an initial solution.\
     Contrary to CLS, CFS uses an exponential zoom on the best solution found, it works at a much smaller scale than the CLS.
    It is continuous optimization, so the :ref:`sp` is converted to continuous.
    To do so, it uses a :ref:`cmap`, such as Henon or Kent map.

    Attributes
    ----------
    level : int
        Chaotic level corresponds to the number of iteration of the chaotic map
    map : Chaos_map
        Chaotic map used to sample points. See Chaos_map object.
    polygon : int
        Vertex number of the rotating polygon (has an influence on the number of evaluated points)
    red_rate : float
        Reduction rate of the progressive zoom on the best solution found
    up_bounds : list
        List of float containing the upper bounds of the search space converted to continuous
    lo_bounds : list
        List of float containing the lower bounds of the search space converted to continuous
    center : float
        List of floats containing the coordinates of the search space center converted to continuous
    radius : float
        List of floats containing the radius for each dimensions of the search space converted to continuous

    See Also
    --------
    :ref:`meta` : Parent class defining what a Metaheuristic is
    Chaotic_optimization : CLS is used here to perform an exploitation
    CGS : Chaotic Global Search
    CLS : Chaotic Local Search

    Examples
    --------
    >>> from zellij.core.loss_func import Loss
    >>> from zellij.core.search_space import Searchspace
    >>> from zellij.strategies.chaos_algorithm import CFS
    >>> from zellij.strategies.utils.chaos_map import Henon
    >>> from zellij.utils.benchmark import himmelblau
    ...
    >>> labels = ["a","b","c"]
    >>> types = ["R","R","R"]
    >>> values = [[-5, 5],[-5, 5],[-5, 5]]
    >>> sp = Searchspace(labels,types,values)
    >>> lf = Loss()(himmelblau)
    ...
    ...          # Henon(map size, dimensions)
    >>> chaosmap = Henon(50,3)
    ...                  # 2xpolygon points/iterations: 2x10x50=1000
    >>> cfs = CFS(lf, sp, 1000, 50, 10, chaosmap)
    >>> point = sp.random_point()[0]
    >>> cfs.run(point, lf([point])[0])
    >>> cfs.show()

    """

    def __init__(
        self,
        loss_func,
        search_space,
        f_calls,
        level,
        polygon,
        map,
        verbose=True,
    ):

        """__init__(loss_func, search_space, f_calls, level, polygon, map, red_rate=0.5, verbose=True)

        Initialize CLS class

        Parameters
        ----------
        loss_func : Loss
            Loss function to optimize. must be of type f(x)=y
        search_space : Searchspace
            Search space object containing bounds of the search space
        f_calls : int
            Maximum number of loss_func calls
        level : int
            Chaotic level corresponds to the number of iteration of the chaotic map
        polygon : int
            Vertex number of the rotating polygon (has an influence on the number of evaluated points)
        map : Chaos_map
            Chaotic map used to sample points. See Chaos_map object.
        verbose : boolean, default=True
            Algorithm verbosity

        """

        ##############
        # PARAMETERS #
        ##############

        super().__init__(loss_func, search_space, f_calls, verbose)

        self.level = level
        self.polygon = polygon
        self.map = map.map
        self.red_rate = np.random.random()

        #############
        # VARIABLES #
        #############

        self.up_bounds = np.array([1 for _ in self.search_space.values])
        self.lo_bounds = np.array([0 for _ in self.search_space.values])

        self.up_plus_lo = self.up_bounds + self.lo_bounds
        self.up_m_lo = self.up_bounds - self.lo_bounds

        self.center = np.multiply(0.5, self.up_plus_lo)
        self.radius = np.multiply(0.5, self.up_m_lo)
        self.center_m_lo_bounds = self.center - self.lo_bounds

        trigo_val = 2 * np.pi / self.polygon
        self.H = [np.zeros(self.polygon), np.zeros(self.polygon)]

        for i in range(1, self.polygon + 1):
            # Initialize trigonometric part of symetric variables (CLS & CFS)
            self.H[0][i - 1] = np.cos(trigo_val * i)
            self.H[1][i - 1] = np.sin(trigo_val * i)

    def stochastic_round(self, solution, k):
        print(solution)
        r = np.random.uniform(-1, 1, len(solution))
        # perturbation on CFS zoom
        z = np.round(solution.astype(float)) + (k % 2) * r

        return z

    def run(
        self, X0=None, Y0=None, chaos_level=0, shift=1, H=None, n_process=1
    ):

        """run(X0=None, Y0=None, chaos_level=0, shift=1, H=None, n_process=1)

        Parameters
        ----------
        X0 : list[float], optional
            Initial solution. If None, a Fractal must be given (H!=None)
        Y0 : {int, float}, optional
            Score of the initial solution
        chaos_level : int, default=0
            Determines at which level of the chaos map, the algorithm starts
        shift : int, default=1
            Determines the starting point of the chaotic map.
        H : Fractal, optional
            When used by FDA, a fractal corresponding to the current subspace is given
        n_process : int, default=1
            Determines the number of best solution found to return.

        Returns
        -------
        best_sol : list[float]
            Returns a list of the <n_process> best found points to the continuous format

        best_scores : list[float]
            Returns a list of the <n_process> best found scores associated to best_sol

        """

        logger.info("CFS starting")

        self.build_bar(self.level)

        if X0:
            self.X0 = np.array(
                self.search_space.convert_to_continuous([X0])[0], dtype=float
            )
        elif H:
            self.X0 = H.center
        else:
            raise ValueError("No starting point given to Simulated Annealing")

        if Y0:
            self.Y0 = Y0
        else:
            logger.info("CLS evaluating initial solution")
            self.Y0 = self.loss_func(
                self.search_space.convert_to_continuous([self.X0], True)
            )[0]

        self.k = shift
        self.chaos_level = chaos_level

        shift = self.chaos_level * (self.k - 1) * self.level

        y = self.map[shift]
        # Limits of the search area, if parameter greater than center, then = 1 else = -1, used to avoid overflow
        db = np.minimum(self.up_bounds - self.X0, self.X0 - self.lo_bounds)

        r_g = np.zeros(self.search_space.n_variables)

        # Randomly select the reduction rate
        # red_rate = random.random()*0.5

        xc = self.X0
        zc = self.Y0

        center_m_solution = self.center - self.X0
        points = np.empty((0, self.search_space.n_variables), dtype=float)

        n_points = self.loss_func.calls
        l = 0

        logger.info("CFS computing chaotic points")

        # for each level of chaos
        while l < self.level and n_points < self.f_calls:

            # Local search area radius
            self.red_rate = np.random.random()
            Rl = self.radius * self.red_rate

            # Decomposition vector
            d = np.random.randint(self.search_space.n_variables)

            # Exponential Zoom factor on the search window
            pc = 10 ** (l + 1)

            # Compute the error/the perturbation applied to the solution
            error_g = np.absolute(
                self.X0 - (self.stochastic_round(pc * self.X0, shift + l) / pc)
            )

            r = np.random.random()

            # for each parameter of a solution determines the improved radius
            r_g = np.minimum((Rl * error_g) / (l ** 2 + 1), db)

            # Compute both chaotic variable of the polygonal model thanks to a chaotic map
            xv = [np.multiply(r_g, y), np.multiply(r_g, y)]

            # For both chaotic variable
            for x in xv:
                xi = np.outer(self.H[1], x)
                xi[:, d] = x[d] * self.H[0]
                xt = self.X0 + xi

                points = np.append(points, xt, axis=0)
                n_points += self.polygon

            l += 1
            self.meta_pb.update()

        # Update progress bar
        self.pending_pb(len(points))

        logger.info("CFS evaluating chaotic points")
        ys = self.loss_func(
            self.search_space.convert_to_continuous(points, True),
            algorithm="CFS",
        )

        # Update progress bar
        self.update_main_pb(
            len(points), explor=True, best=self.loss_func.new_best
        )

        ys = np.array(ys)
        idx = np.array(np.argsort(ys))[:n_process]

        # best solution found
        best_sol = points[idx]
        best_scores = ys[idx]

        self.close_bar()

        logger.info("CFS ending")

        return best_sol, best_scores


class Chaotic_optimization(Metaheuristic):

    """Chaotic_optimization

    Chaotic optimization combines CGS, CLS and CFS. Using a unique chaos map. You can determine the number of outer and inner iteration is determine using an exploration ratio,\
     and according to chaotic levels associated to CGS, CLS and CFS. The best solution found by CGS is used as a starting for CLS, and the best solution found by CLS is used by CFS.

    Attributes
    ----------

    chaos_map : {'henon', 'kent', 'tent', 'logistic', 'random', Chaos_map}
        If a string is given, the algorithm will select the corresponding map. The chaotic map is used to sample points.\
         If it is a map, it will directly use it. Be carefull, the map size must be sufficient according to the parametrization.

    exploration_ratio : float
        It will determine the number of calls to the loss function dedicated to exploration and exploitation, according to chaotic levels associated to CGS, CLS and CFS.

    polygon : int
        Vertex number of the rotating polygon (has an influence on the number of evaluated points) for CLS and CFS

    red_rate : float
        Reduction rate of the progressive zoom on the best solution found for CLS and CFS

    CGS_level : int
        Number of chaotic level associated to CGS

    CLS_level : int
        Number of chaotic level associated to CLS

    CFS_level : int
        Number of chaotic level associated to CFS

    verbose : boolean, default=True
        Algorithm verbosity

    Methods
    -------

    run(self, n_process=1)
        Runs Chaotic_optimization

    show(filename=None)
        Plots results

    See Also
    --------
    :ref:`meta` : Parent class defining what a Metaheuristic is
    CGS : Chaotic Global Search
    CLS : Chaotic Local Search
    CFS : Chaotic Fine Search
    """

    def __init__(
        self,
        loss_func,
        search_space,
        f_calls,
        chaos_map="henon",
        exploration_ratio=0.30,
        levels=(32, 6, 2),
        polygon=4,
        red_rate=0.5,
        verbose=True,
    ):

        """__init__(loss_func, search_space, f_calls,chaos_map="henon", exploration_ratio = 0.70, levels = (32,6,2), polygon=4, red_rate=0.5, verbose=True)

        Initialize CGS class

        Parameters
        ----------
        loss_func : Loss
            Loss function to optimize. must be of type f(x)=y
        search_space : Searchspace
            Search space object containing bounds of the search space
        f_calls : int
            Maximum number of loss_func calls
        chaos_map : {'henon', 'kent', 'tent', 'logistic', 'random', Chaos_map}
            If a string is given, the algorithm will select the corresponding map. The chaotic map is used to sample points.\
             If it is a map, it will directly use it. Be carefull, the map size must be sufficient according to the parametrization.
        exploration_ratio : float, default=0.80
            Must be between 0 and 1.\
            It will determine the number of calls to the loss function dedicated to exploration and exploitation, according to chaotic levels associated to CGS, CLS and CFS.
        levels : (int, int, int)
            Used to determine the number of chaotic levels for respectively, CGS, CLS and CFS.
        polygon : int, default=4
            Vertex number of the rotating polygon (has an influence on the number of evaluated points) for CLS and CFS
        red_rate : float, default=0.5
            Reduction rate of the progressive zoom on the best solution found
        verbose : boolean, default=True
            Algorithm verbosity

        """

        ##############
        # PARAMETERS #
        ##############

        super().__init__(loss_func, search_space, f_calls, verbose)

        self.chaos_map = chaos_map
        self.exploration_ratio = exploration_ratio
        self.polygon = polygon
        self.red_rate = red_rate

        self.CGS_level = levels[0]
        self.CLS_level = levels[1]
        self.CFS_level = levels[2]

        #############
        # VARIABLES #
        #############

        if self.CGS_level > 0:
            if self.CLS_level != 0 or self.CFS_level != 0:
                self.iterations = np.ceil(
                    (self.f_calls * self.exploration_ratio)
                    / (4 * self.CGS_level)
                )
                self.inner_iterations = np.ceil(
                    (self.f_calls * (1 - self.exploration_ratio))
                    / (
                        (self.CLS_level + self.CFS_level)
                        * self.polygon
                        * self.iterations
                    )
                )
            else:
                self.iterations = np.ceil(self.f_calls / (4 * self.CGS_level))
                self.inner_iterations = 0
        else:
            raise ValueError("CGS level must be > 0")

        if type(chaos_map) == str:
            self.map_size = int(
                np.max(
                    [
                        self.iterations * self.CGS_level,
                        self.iterations
                        * self.inner_iterations
                        * self.CLS_level,
                        self.iterations
                        * self.inner_iterations
                        * self.CFS_level,
                    ]
                )
            )
        else:
            self.map_size = int(
                np.ceil(
                    np.max(
                        [
                            self.iterations * self.CGS_level,
                            self.iterations
                            * self.inner_iterations
                            * self.CLS_level,
                            self.iterations
                            * self.inner_iterations
                            * self.CFS_level,
                        ]
                    )
                    / len(self.chaos_map)
                )
            )

        self.map = chaos_map_name[self.chaos_map](
            self.map_size, self.search_space.n_variables
        )

        logging.info(str(self))

    def run(self, H=None, n_process=1):

        """run(H=None, n_process=1)

        Runs the Chaotic_optimization

        Parameters
        ----------
        H : Fractal, default=None
            When used by FDA, a fractal corresponding to the current subspace is given

        n_process : int, default=1
            Determine the number of best solution found to return.

        Returns
        -------
        best_sol : list[float]
            Returns a list of the <n_process> best found points to the continuous format

        best_scores : list[float]
            Returns a list of the <n_process> best found scores associated to best_sol

        """

        logger.info("Chaotic optimization starting")

        # Progress bar
        self.build_bar(self.iterations * self.inner_iterations)

        # Initialize CGS/CLS/CFS
        cgs = CGS(
            self.loss_func,
            self.search_space,
            self.f_calls,
            self.CGS_level,
            self.map,
            verbose=self.verbose,
        )
        cls = CLS(
            self.loss_func,
            self.search_space,
            self.f_calls,
            self.CLS_level,
            self.polygon,
            self.map,
            self.red_rate,
            verbose=self.verbose,
        )
        cfs = CFS(
            self.loss_func,
            self.search_space,
            self.f_calls,
            self.CFS_level,
            self.polygon,
            self.map,
            self.red_rate,
            verbose=self.verbose,
        )

        cgs.manager, cls.manager, cfs.manager = (
            self.manager,
            self.manager,
            self.manager,
        )

        # Initialize historic vector
        best_sol = np.array([])
        best_scores = np.array([])

        k = 1

        # Outer loop (exploration)
        while k <= self.iterations and self.loss_func.calls < self.f_calls:

            logger.info("Chaotic optimization: Exploration phase")

            # If there is CGS
            if self.CGS_level > 0:

                prec_calls = self.loss_func.calls
                self.pending_pb(self.CGS_level * 4)

                x_inter, loss_value = cgs.run(k)

                self.update_main_pb(
                    self.loss_func.calls - prec_calls,
                    explor=True,
                    best=self.loss_func.new_best,
                )

                # Store to return best solution found
                best_sol = np.append(best_sol, x_inter)
                best_scores = np.append(best_scores, loss_value)

            # Else select random point for the exploitation
            else:

                logger.warning(
                    "Chaotic optimization: using random instead of CGS"
                )

                x_inter = [np.random.random(self.search_space.n_variables)]

                self.pending_pb(1)

                loss_value = self.loss_func(x_inter)

                self.update_main_pb(
                    1, explor=True, best=self.loss_func.new_best
                )

                # Store to return best solution found
                best_sol = np.append(x_inter)
                best_scores = np.append(loss_value)

            logger.debug(
                f"Iterations | Loss function calls | Best value from CGS"
            )
            logger.debug(
                f"{k} < {self.iterations} | {self.loss_func.calls} < {self.f_calls} | {loss_value}"
            )
            logger.debug(f"New best solution found {self.loss_func.new_best}")

            inner = 0

            # Inner loop (exploitation)
            while (
                inner < self.inner_iterations
                and self.loss_func.calls < self.f_calls
            ):

                logger.info("Chaotic optimization: Exploitation phase")

                if self.CLS_level > 0:

                    prec_calls = self.loss_func.calls
                    self.pending_pb(self.CLS_level * self.polygon * 2)

                    x_inter, loss_value = cls.run(
                        x_inter[0], loss_value[0], inner, k
                    )

                    self.update_main_pb(
                        self.loss_func.calls - prec_calls,
                        explor=False,
                        best=self.loss_func.new_best,
                    )

                    # Store to return best solution found
                    best_sol = np.append(best_sol, x_inter)
                    best_scores = np.append(best_scores, loss_value)

                if self.CFS_level > 0:

                    prec_calls = self.loss_func.calls
                    self.pending_pb(self.CFS_level * self.polygon * 2)

                    x_inter, loss_value = cfs.run(
                        x_inter[0], loss_value[0], inner, k
                    )

                    self.update_main_pb(
                        self.loss_func.calls - prec_calls,
                        explor=False,
                        best=self.loss_func.new_best,
                    )

                    # Store to return best solution found
                    best_sol = np.append(best_sol, x_inter)
                    best_scores = np.append(best_scores, loss_value)

                logger.debug(
                    f"Iterations | Loss function calls | Best value from CGS"
                )
                logger.debug(
                    f"{k} < {self.iterations} | {self.loss_func.calls} < {self.f_calls} | {loss_value}"
                )
                logger.debug(
                    f"New best solution found {self.loss_func.new_best}"
                )

                inner += 1

                self.meta_pb.update()

            ind_min = np.argsort(best_scores)[0:n_process]
            best_scores = np.array(best_scores)[ind_min].tolist()
            best_sol = np.array(best_sol)[ind_min].tolist()

            k += 1

        self.close_bar()

        logger.info("Chaotic optimization ending")

        return best_sol, best_scores

    def show(self, filepath="", save=False):

        """show(filename="")

        Plots solutions and scores evaluated during the optimization

        Parameters
        ----------
        filename : str, default=None
            If a filepath is given, the method will read the file and will try to plot contents.

        save : boolean, default=False
            Save figures
        """

        super().show(filepath, save)

    def __str__(self):
        return f"Max Loss function calls:{self.f_calls}\nDimensions:{self.search_space.n_variables}\nExploration/Exploitation:{self.exploration_ratio}|{1-self.exploration_ratio}\nRegular polygon:{self.polygon}\nZoom:{self.red_rate}\nIterations:\n\tGlobal:{self.iterations}\n\tInner:{self.inner_iterations}\nChaos Levels:\n\tCGS:{self.CGS_level}\n\tCLS:{self.CLS_level}\n\tCFS:{self.CFS_level}\nMap size:{self.map_size}x{self.search_space.n_variables}"



# @Author: Thomas Firmin <ThomasFirmin>
# @Date:   2022-05-03T15:41:48+02:00
# @Email:  thomas.firmin@univ-lille.fr
# @Project: Zellij
# @Last modified by:   ThomasFirmin
# @Last modified time: 2022-05-03T15:45:06+02:00
# @License: CeCILL-C (http://www.cecill.info/index.fr.html)
# @Copyright: Copyright (C) 2022 Thomas Firmin


from zellij.core.metaheuristic import Metaheuristic
from deap import base
from deap import creator
from deap import tools
import numpy as np
import os
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns

import logging

logger = logging.getLogger("zellij.GA")


class Genetic_algorithm(Metaheuristic):

    """Genetic_algorithm

    Genetic_algorithm (GA) implements a steady state genetic algorithm. It can be used for exploration and exploitation.
    Indeed when the population has converged, GA can ,thanks to the mutation and crossover operators, perform an intensification phase arround best solutions.
    It can work with a mixed search space, by adapting its operator.

    **Will be modified soon:**
    Used operators are: One-point crossover and Tournament selection of size 3.

    Here the mutation operator is the neighborhood defined in the :ref:`sp` object.
    Available crossover operator are those compatible with a mixed individual (
    `1-point <https://deap.readthedocs.io/en/master/api/tools.html#deap.tools.cxOnePoint>`_,
    `2-points <https://deap.readthedocs.io/en/master/api/tools.html#deap.tools.cxTwoPoint>`_
    ...).
    Same with the slection (
    `Roulette <https://deap.readthedocs.io/en/master/api/tools.html#deap.tools.selRoulette>`_,
    `Tournament <https://deap.readthedocs.io/en/master/api/tools.html#deap.tools.selTournament>`_
    ).

    It uses `DEAP <https://deap.readthedocs.io/>`_.
    See :ref:`meta` for more info.

    Attributes
    ----------

    pop_size : int
        Population size of the GA.\
        In a distributed environment (e.g. MPILoss), it has an influence on the parallelization quality.\
        It must be tuned according the available hardware.

    generation : int
        Generation number of the GA.

    elitism : float, default=0.5
        Percentage of the best parents to keep in the next population by replacing the worst children.
        Default 50%.

    filename : str, optional
        If a file containing initial solutions. GA will initialize the population with it.


    See Also
    --------
    :ref:`meta` : Parent class defining what a Metaheuristic is in Zellij.
    :ref:`lf` : Describes what a loss function is in Zellij.
    :ref:`sp` : Describes what a search space is in Zellij.

    Examples
    --------
    >>> from zellij.core.loss_func import Loss
    >>> from zellij.core.search_space import Searchspace
    >>> from zellij.strategies.genetic_algorithm import Genetic_algorithm
    >>> from zellij.utils.benchmark import himmelblau
    >>> labels = ["a","b","c"]
    >>> types = ["R","R","R"]
    >>> values = [[-5, 5],[-5, 5],[-5, 5]]
    >>> sp = Searchspace(labels,types,values)
    >>> lf = Loss()(himmelblau)
    >>> ga = Genetic_algorithm(lf, sp, 1000, pop_size=25, generation=40,elitism=0.5)
    >>> ga.run()
    >>> ga.show()


    .. image:: ../_static/ga_sp_ex.png
        :width: 924px
        :align: center
        :height: 487px
        :alt: alternate text
    .. image:: ../_static/ga_res1_ex.png
        :width: 924px
        :align: center
        :height: 487px
        :alt: alternate text
    .. image:: ../_static/ga_res2_ex.png
        :width: 924px
        :align: center
        :height: 487px
        :alt: alternate text
    """

    def __init__(
        self,
        loss_func,
        search_space,
        f_calls,
        pop_size=10,
        generation=1000,
        elitism=0.5,
        filename="",
        verbose=True,
    ):

        """__init__(loss_func, search_space, f_calls, pop_size = 10, generation = 1000, verbose=True)

        Initialize Genetic_algorithm class

        Parameters
        ----------
        loss_func : Loss
            Loss function to optimize. must be of type f(x)=y

        search_space : Searchspace
            Search space object containing bounds of the search space.

        f_calls : int
            Maximum number of loss_func calls

        pop_size : int
            Population size of the GA.\
            In a distributed environment (e.g. MPILoss), it has an influence on the parallelization quality.\
            It must be tuned according the available hardware.

        generation : int
            Generation number of the GA.

        elitism : float
            Percentage of the best parents to keep in the next population by replacing the worst children.

        filename : str, optional
            If a file containing initial solutions. GA will initialize the population with it.

        verbose : boolean, default=True
            Algorithm verbosity

        """

        super().__init__(loss_func, search_space, f_calls, verbose)

        self.pop_size = pop_size
        self.generation = generation
        self.elitism = elitism

        self.pop_historic = []
        self.fitness_historic = []

        # Population save
        self.filename = filename
        self.ga_save = ""

    # Define what an individual is
    def define_individual(self):
        """define_individual(self)

        Describe how an individual should be initialized. Here a random point from SearchSpace is sampled.

        """
        # Select one random point from the search space
        solution = self.search_space.random_point()[0]

        return solution

    # Initialize an individual extracted from a file
    def initIndividual(self, icls, content):
        """initIndividual(self, icls, content)

        Initialize an individual to DEAP.

        """
        return icls([content.to_list()])

    # Initialize a population extracted from a file
    def initPopulation(self, pcls, ind_init, filename):
        """initPopulation(self, pcls, ind_init, filename)

        Initialize a population of individual, from a file, to DEAP.

        """
        data = pd.read_csv(
            filename,
            sep=",",
            decimal=".",
            usecols=self.search_space.n_variables,
        )
        contents = data.tail(taille_population)

        return pcls(ind_init(c) for index, c in contents.iterrows())

    # Mutate operator
    def mutate(self, individual, proba):

        """mutate(self, individual, proba)

        Mutate a given individual, using Searchspace neighborhood.

        Parameters
        ----------
        individual : list[{int, float, str}]
            Individual to mutate, in the mixed format.

        proba : float
            Probability to mutate a gene.

        Returns
        -------

        individual : list[{int, float, str}]
            Mutated individual

        """

        # For each dimension of a solution draw a probability to be muted
        for index, label in enumerate(self.search_space.labels):
            t = np.random.random()
            if t < proba:

                # Get a neighbor of the selected attribute
                individual[0][index] = self.search_space.get_neighbor(
                    individual[0], attribute=label
                )[0]

        return (individual,)

    # Run GA
    def run(self, H=None, n_process=1):

        """run(H=None, n_process=1)

        Runs GA

        Parameters
        ----------
        H : Fractal, optional
            When used by FDA, a fractal corresponding to the current subspace is given

        n_process : int, default=1
            Determine the number of best solution found to return.

        Returns
        -------
        best_sol : list[float]
            Returns a list of the <n_process> best found points to the continuous format

        best_scores : list[float]
            Returns a list of the <n_process> best found scores associated to best_sol

        """

        # Progress bar
        self.build_bar(self.generation)

        self.loss_func.file_created = False

        logger.info("Starting")

        logger.info("Constructing tools...")

        # Define problem type "fitness", weights = -1.0 -> minimization problem
        creator.create("FitnessMin", base.Fitness, weights=(-1.0,))

        # Define what an individual is for the algorithm
        creator.create("Individual", list, fitness=creator.FitnessMin)

        # Toolbox contains all the operator of GA. (mutate, select, crossover...)
        toolbox = base.Toolbox()

        # Start from a saved population
        if self.filename:

            toolbox.register(
                "individual_guess", self.initIndividual, creator.Individual
            )
            toolbox.register(
                "population_guess",
                self.initPopulation,
                list,
                toolbox.individual_guess,
                self.filename,
            )

            logger.info("Creation of the initial population...")
            pop = toolbox.population_guess()

        # Start from a random population
        else:
            # Determine what is an individual
            toolbox.register("hyperparameters", self.define_individual)

            # Determine the way to build individuals for the population
            toolbox.register(
                "individual",
                tools.initRepeat,
                creator.Individual,
                toolbox.hyperparameters,
                n=1,
            )

            # Determine the way to build a population
            toolbox.register(
                "population", tools.initRepeat, list, toolbox.individual
            )

            logger.info("Creation of the initial population...")

            # Build the population
            pop = toolbox.population(n=self.pop_size)

        # Create crossover tool
        toolbox.register("mate", tools.cxOnePoint)
        # Create mutation tool
        toolbox.register(
            "mutate", self.mutate, proba=1 / self.search_space.n_variables
        )
        # Create selection tool
        toolbox.register("select", tools.selTournament, tournsize=3)

        # Create a tool to select best individuals from a population
        bpn = int(self.pop_size * self.elitism)
        bcn = self.pop_size - bpn
        toolbox.register("best_p", tools.selBest, k=bpn)
        toolbox.register("best_c", tools.selBest, k=bcn)

        best_of_all = tools.HallOfFame(n_process)

        # Ga initialization

        logger.info("Evaluating the initial population...")
        # Compute dynamically fitnesses
        solutions = []
        solutions = [p[0] for p in pop]

        # Progress bar
        self.pending_pb(len(solutions))

        fitnesses = self.loss_func(solutions, generation=0)

        self.update_main_pb(
            len(solutions), explor=True, best=self.loss_func.new_best
        )

        # Map computed fitness to individual fitness value
        for ind, fit in zip(pop, fitnesses):
            ind.fitness.values = (fit,)

        fits = [ind.fitness.values[0] for ind in pop]

        # Save file
        if self.loss_func.save:
            self.ga_save = os.path.join(
                self.loss_func.outputs_path, "ga_population.csv"
            )
            with open(self.ga_save, "a") as f:
                f.write(
                    ",".join(e for e in self.search_space.labels) + ",loss\n"
                )
                for ind, cout in zip(pop, fits):
                    f.write(
                        ",".join(str(e) for e in ind[0])
                        + ","
                        + str(cout)
                        + "\n"
                    )

        for ind, cout in zip(pop, fits):
            self.pop_historic.append(ind[0])
            self.fitness_historic.append(cout)

        logger.info("Initial population evaluated")

        logger.info("Evolution starting...")
        g = 0
        while g < self.generation and self.loss_func.calls < self.f_calls:
            g += 1

            # Progress bar
            self.meta_pb.update()

            # Update all of fame
            best_of_all.update(pop)

            if self.verbose:
                logger.debug("Generation: " + str(g))

                # Selection operator
                logger.debug("Selection...")

            offspring = toolbox.select(pop, k=len(pop))

            offspring = list(map(toolbox.clone, offspring))

            children = []

            # Crossover operator
            if self.verbose:
                logger.debug("Crossover...")

            i = 0
            for child1, child2 in zip(offspring[::2], offspring[1::2]):

                # Clone individuals from crossover
                children1 = toolbox.clone(child1)
                children2 = toolbox.clone(child2)

                # Apply crossover
                toolbox.mate(children1[0], children2[0])
                # Delete children fitness inherited from the parents
                del children1.fitness.values
                del children2.fitness.values

                # Add new children to list
                children.append(children1)
                children.append(children2)

            # Mutate children
            if self.verbose:
                logger.debug("Mutation...")
            for mutant in children:
                toolbox.mutate(mutant)

            if self.verbose:
                logger.debug("Evaluating population n°" + str(g))

            # Compute dynamically fitnesses
            solutions = []
            solutions = [p[0] for p in children]

            # progress bar
            self.pending_pb(len(solutions))

            fitnesses = self.loss_func(solutions, generation=g)

            # Progress bar
            self.update_main_pb(
                len(solutions), explor=True, best=self.loss_func.new_best
            )

            # Map computed fitness to individual fitness value
            for ind, fit in zip(children, fitnesses):
                ind.fitness.values = (fit,)

            # Build new population
            pop[:] = toolbox.best_p(offspring) + toolbox.best_c(children)

            # Get fitnesses from the new population
            fits = [ind.fitness.values[0] for ind in pop]

            # Save new population
            if self.loss_func.save:
                with open(self.ga_save, "a") as f:
                    for ind, cout in zip(pop, fits):
                        f.write(
                            ",".join(str(e) for e in ind[0])
                            + ","
                            + str(cout)
                            + "\n"
                        )

            for ind, cout in zip(pop, fits):
                self.pop_historic.append(ind[0])
                self.fitness_historic.append(cout)

            # End population evaluation
            logger.info(f"Evaluation n°{g} ending...")

        best = []
        min = []

        logger.info("Ending")
        for b in best_of_all:
            min.append(b.fitness.values[0])
            best.append(b[0])

            # print best parameters from genetic algorithm
            logger.info(
                "Best parameters: "
                + str(b[0])
                + " | score: "
                + str(b.fitness.values[0])
            )

        self.close_bar()
        return best, min

    def show(self, filepath="", save=False):

        """show(filepath="")

        Plots solutions and scores evaluated during the optimization

        Parameters
        ----------
        filename : str, default=None
            If a filepath is given, the method will read the file and will try to plot contents.

        save : boolean, default=False
            Save figures
        """

        all_data, all_scores = super().show(filepath, save)

        if filepath:
            gapth = os.path.join(filepath, "outputs", "ga_population.csv")
            data = pd.read_table(gapth, sep=",", decimal=".")
            scores = data["loss"].to_numpy()
        else:
            data = self.pop_historic
            scores = np.array(self.fitness_historic)

        quantile = np.quantile(scores, 0.75)
        argmin = np.argmin(scores)
        min = scores[argmin]

        # Padding missing individual for reshape
        m, n = int(np.ceil(len(scores) / self.pop_size)), self.pop_size
        scores = np.pad(
            scores,
            (0, m * n - scores.size),
            mode="constant",
            constant_values=np.nan,
        )

        heatmap = scores.reshape((m, n))

        minimums = np.min(heatmap, axis=1)
        means = np.mean(heatmap, axis=1)

        heatmap.sort(axis=1)
        heatmap = heatmap.transpose()

        fig, ax = plt.subplots(figsize=(19.2, 14.4))
        ax = sns.heatmap(
            heatmap,
            vmin=min,
            vmax=quantile,
            cmap="YlGnBu",
            cbar_kws={"label": "Fitness"},
        )
        ax.invert_yaxis()
        ax.set_title(f"Fitness of each individual through generations")
        ax.set(xlabel="Generation number", ylabel="Individual number")

        if save:
            save_path = os.path.join(
                self.loss_func.plots_path, f"heatmap_ga.png"
            )
            plt.savefig(save_path, bbox_inches="tight")
            plt.close()
        else:
            plt.show()
            plt.close()

        fig, ax = plt.subplots(figsize=(19.2, 14.4))
        plt.plot(
            np.arange(len(minimums)),
            minimums,
            "-",
            label="Best individual",
            color="red",
        )
        plt.plot(np.arange(len(means)), means, ":", label="Mean", color="blue")
        plt.title("Best individual and population's mean through generations")
        plt.xlabel("Generations")
        plt.ylabel("Score")
        plt.legend()

        if save:
            save_path = os.path.join(
                self.loss_func.plots_path, f"lineplot_ga.png"
            )

            plt.savefig(save_path, bbox_inches="tight")
            plt.close()
        else:
            plt.show()
            plt.close()



# @Author: Thomas Firmin <ThomasFirmin>
# @Date:   2022-05-03T15:41:48+02:00
# @Email:  thomas.firmin@univ-lille.fr
# @Project: Zellij
# @Last modified by:   ThomasFirmin
# @Last modified time: 2022-05-03T15:45:01+02:00
# @License: CeCILL-C (http://www.cecill.info/index.fr.html)
# @Copyright: Copyright (C) 2022 Thomas Firmin


from zellij.core.metaheuristic import Metaheuristic
import matplotlib.pyplot as plt

import torch
import numpy as np
from botorch.models import SingleTaskGP
from gpytorch.mlls.sum_marginal_log_likelihood import ExactMarginalLogLikelihood
from botorch.optim import optimize_acqf
from botorch.acquisition.analytic import ExpectedImprovement
from botorch import fit_gpytorch_model
import time

import logging

logger = logging.getLogger("zellij.BO")


class Bayesian_optimization(Metaheuristic):
    """Bayesian_optimization

    Bayesian optimization (BO) is a surrogate based optimization method which
    interpolates the actual loss function with a surrogate model, here it is a
    gaussian process. By sampling into this surrogate using a metaheuristic,
    BO determines promising points, which are worth to evaluate with the actual
    loss function. Once done, the gaussian process is updated using results
    obtained by evaluating those encouraging solutions with the loss function.

    It is based on `BoTorch <https://botorch.org/>`_ and `GPyTorch <https://gpytorch.ai/>`_.

    Attributes
    ----------
    loss_func : Loss
        Loss function to optimize. must be of type f(x)=y
    search_space : Searchspace
        Search space object containing bounds of the search space
    f_calls : int
        Maximum number of loss_func calls
    verbose : bool
        If False, there will be no print and no progress bar.
    surrogate : botorch.models.model.Model, default=SingleTaskGP
        Gaussian Process Regressor object from 'botorch'.
        Determines the surrogate model that Bayesian optimization will use to
        interpolate the loss function
    likelihood : gpytorch.mlls, default=ExactMarginalLogLikelihood
        gpytorch.mlls object it determines which MarginalLogLikelihood to use
        when optimizing kernel's hyperparameters
    acquisition : botorch.acquisition.acquisition.AcquisitionFunction, default = ExpectedImprovement
        An acquisition function or infill criteria, determines how 'promising'
        a point sampled from the surrogate is.
    initial_size : int, default=10
        Size of the initial set of solution to draw randomly.
    sampler : botorch.sampling.samplers, default=None
        Sampler used for a full Bayesian approach with Monte-Carlo sampling
        applied to approximate the integrated acquisition function.

    gpu: bool, default=True
        Use GPU if available

    See Also
    --------
    :ref:`meta` : Parent class defining what a Metaheuristic is
    :ref:`lf` : Describes what a loss function is in Zellij
    :ref:`sp` : Describes what a loss function is in Zellij

    Examples
    --------
    >>> from zellij.core.loss_func import Loss
    >>> from zellij.core.search_space import Searchspace
    >>> from zellij.utils.benchmark import himmelblau
    >>> from zellij.strategies.bayesian_optimization import Bayesian_optimization
    >>> import botorch
    >>> import gpytorch
    ...
    >>> labels = ["a","b","c"]
    >>> types = ["R","R","R"]
    >>> values = [[-5, 5],[-5, 5],[-5, 5]]
    >>> sp = Searchspace(labels,types,values)
    >>> lf = Loss()(himmelblau)
    ...
    >>> bo = Bayesian_optimization(lf, sp, 500,
    ...       acquisition=botorch.acquisition.monte_carlo.qExpectedImprovement,
    ...       q=5)
    >>> bo.run()
    >>> bo.show()


    .. image:: ../_static/bo_sp_ex.png
        :width: 924px
        :align: center
        :height: 487px
        :alt: alternate text
    .. image:: ../_static/bo_res_ex.png
        :width: 924px
        :align: center
        :height: 487px
        :alt: alternate text

    """

    def __init__(
        self,
        loss_func,
        search_space,
        f_calls,
        verbose=True,
        surrogate=SingleTaskGP,
        likelihood=ExactMarginalLogLikelihood,
        acquisition=ExpectedImprovement,
        initial_size=10,
        sampler=None,
        gpu=True,
        **kwargs,
    ):
        """Short summary.

        Parameters
        ----------
        loss_func : Loss
            Loss function to optimize. must be of type f(x)=y
        search_space : Searchspace
            Search space object containing bounds of the search space
        f_calls : int
            Maximum number of loss_func calls
        verbose : bool
            If False, there will be no print and no progress bar.
        surrogate : botorch.models.model.Model, default=SingleTaskGP
            Gaussian Process Regressor object from 'botorch'.
            Determines the surrogate model that Bayesian optimization will use to
            interpolate the loss function
        likelihood : gpytorch.mlls, default=ExactMarginalLogLikelihood
            gpytorch.mlls object it determines which MarginalLogLikelihood to use
            when optimizing kernel's hyperparameters
        acquisition : botorch.acquisition.acquisition.AcquisitionFunction, default = ExpectedImprovement
            An acquisition function or infill criteria, determines how 'promising'
            a point sampled from the surrogate is.
        initial_size : int, default=10
            Size of the initial set of solution to draw randomly.
        sampler : botorch.sampling.samplers, default=None
            Sampler used for a full Bayesian approach with Monte-Carlo sampling
            applied to approximate the integrated acquisition function.
        gpu: bool, default=True
            Use GPU if available
        """

        super().__init__(loss_func, search_space, f_calls, verbose)

        ##############
        # PARAMETERS #
        ##############

        self.acquisition = acquisition
        self.surrogate = surrogate
        self.likelihood = likelihood
        self.initial_size = initial_size
        self.sampler = sampler

        self.kwargs = kwargs

        #############
        # VARIABLES #
        #############
        self.device = torch.device(
            "cuda" if torch.cuda.is_available() else "cpu"
        )
        self.dtype = torch.double

        self.bounds = torch.tensor(
            [
                [0.0] * self.search_space.n_variables,
                [1.0] * self.search_space.n_variables,
            ],
            device=self.device,
            dtype=self.dtype,
        )

        self.best_observed = []

        self.iterations = int(
            np.ceil(
                (self.f_calls - self.initial_size) / self.kwargs.get("q", 1)
            )
        )

    def _generate_initial_data(self):
        # generate training data
        train_x = torch.rand(
            self.initial_size,
            self.search_space.n_variables,
            device=self.device,
            dtype=self.dtype,
        )
        train_obj = -torch.tensor(
            self.loss_func(
                self.search_space.convert_to_continuous(train_x.numpy(), True)
            )
        ).unsqueeze(
            -1
        )  # add output dimension

        return train_x, train_obj, -self.loss_func.best_score

    def _initialize_model(self, train_x, train_obj, state_dict=None):

        # define models for objective and constraint
        model = self.surrogate(
            train_x,
            train_obj,
            **{
                key: value
                for key, value in self.kwargs.items()
                if key in self.surrogate.__init__.__code__.co_varnames
            },
        ).to(train_x)

        mll = self.likelihood(model.likelihood, model)

        # load state dict if it is passed
        if state_dict is not None:
            model.load_state_dict(state_dict)

        return mll, model

    def _optimize_acqf_and_get_observation(
        self, acq_func, restarts=10, raw=512
    ):
        """Optimizes the acquisition function, and returns a new candidate and a noisy observation."""

        # optimize
        candidates, _ = optimize_acqf(
            acq_function=acq_func,
            bounds=self.bounds,
            q=self.kwargs.get("q", 1),
            num_restarts=restarts,
            raw_samples=raw,  # used for intialization heuristic
            options={"batch_limit": 5, "maxiter": 200},
        )

        # observe new values
        new_x = candidates.detach()

        # progress bar
        self.pending_pb(len(new_x))

        new_obj = -torch.tensor(
            self.loss_func(
                self.search_space.convert_to_continuous(new_x.numpy(), True)
            )
        ).unsqueeze(
            -1
        )  # add output dimension

        # progress bar
        self.update_main_pb(
            len(new_x), explor=True, best=self.loss_func.new_best
        )

        return new_x, new_obj

    def run(self, n_process=1):

        """run(n_process=1)

        Runs SA

        Parameters
        ----------
        n_process : int, default=1
            Determine the number of best solution found to return.

        Returns
        -------
        best_sol : list[float]
            Returns a list of the <n_process> best found points to the continuous format

        best_scores : list[float]
            Returns a list of the <n_process> best found scores associated to best_sol

        """

        # progress bar
        self.build_bar(self.iterations)

        self.best_observed = []
        self.best_idx = [0]

        # call helper functions to generate initial training data and initialize model
        (
            train_x,
            train_obj,
            best_observed_value,
        ) = self._generate_initial_data()

        # progress bar
        self.pending_pb(self.initial_size)

        mll, model = self._initialize_model(train_x, train_obj)

        # progress bar
        self.update_main_pb(
            self.initial_size, explor=True, best=self.loss_func.new_best
        )
        self.meta_pb.update()

        self.best_observed.append(self.loss_func.best_score)

        iteration = 1

        # run N_BATCH rounds of BayesOpt after the initial random batch
        while (
            iteration < self.iterations and self.loss_func.calls < self.f_calls
        ):

            # fit the models
            fit_gpytorch_model(mll)

            acqf = self.acquisition(
                model=model,
                best_f=-self.loss_func.best_score,
                sampler=self.sampler,
                X_baseline=(train_x,),
                **{
                    key: value
                    for key, value in self.kwargs.items()
                    if key in self.acquisition.__init__.__code__.co_varnames
                },
            )

            # optimize and get new observation
            (
                new_x,
                new_obj,
            ) = self._optimize_acqf_and_get_observation(acqf)

            # update training points
            train_x = torch.cat([train_x, new_x])
            train_obj = torch.cat([train_obj, new_obj])

            # reinitialize the models so they are ready for fitting on next iteration
            # use the current state dict to speed up fitting
            mll, model = self._initialize_model(
                train_x,
                train_obj,
                model.state_dict(),
            )

            # update progress
            self.best_observed.append(self.loss_func.best_score)
            self.best_idx.append(iteration)

            iteration += 1

            # progress bar
            self.meta_pb.update()

        best_idx = np.argpartition(self.loss_func.all_scores, n_process)
        best = [self.loss_func.all_solutions[i] for i in best_idx[:n_process]]
        min = [self.loss_func.all_scores[i] for i in best_idx[:n_process]]

        # self.close_bar()
        return best, min

    def show(self, filepath=None, save=False):

        """show(self, filename=None)

        Plots solutions and scores computed during the optimization

        Parameters
        ----------
        filepath : str, default=""
            If a filepath is given, the method will read files insidethe folder and will try to plot contents.

        save : boolean, default=False
            Save figures
        """

        data_all, all_scores = super().show(filepath, save)

        min = np.argmin(all_scores)
        indexes = np.repeat(0, self.initial_size)
        indexes = np.append(
            indexes,
            np.repeat(
                np.arange(1, self.iterations, dtype=int),
                self.kwargs.get("q", 1),
            )[: len(all_scores) - self.initial_size],
        )
        plt.scatter(
            indexes,
            all_scores,
            c=all_scores,
            cmap="plasma_r",
        )
        plt.plot(
            self.best_idx,
            self.best_observed,
            color="red",
            lw=2,
            label="Best scores list",
        )
        plt.title("Scores evolution during bayesian optimization")
        plt.scatter(
            min // self.kwargs.get("q", 1),
            all_scores[min],
            color="red",
            label="Best score",
        )
        plt.annotate(
            str(all_scores[min]),
            (min // self.kwargs.get("q", 1), all_scores[min]),
        )
        plt.xlabel("iterations")
        plt.ylabel("Scores")
        plt.legend(loc=1)
        plt.show()



# @Author: Thomas Firmin <ThomasFirmin>
# @Date:   2022-05-03T15:41:48+02:00
# @Email:  thomas.firmin@univ-lille.fr
# @Project: Zellij
# @Last modified by:   ThomasFirmin
# @Last modified time: 2022-05-03T15:45:47+02:00
# @License: CeCILL-C (http://www.cecill.info/index.fr.html)
# @Copyright: Copyright (C) 2022 Thomas Firmin


from zellij.core.metaheuristic import Metaheuristic
from zellij.strategies.utils.cooling import Cooling
import zellij.utils.progress_bar as pb
import numpy as np
import os
import matplotlib.pyplot as plt
import pandas as pd

import logging

logger = logging.getLogger("zellij.SA")


class Simulated_annealing(Metaheuristic):

    """Simulated_annealing

    Simulated_annealing (SA) is an exploitation strategy allowing to do hill climbing by starting from\
    an initial solution and iteratively moving to next one,\
     better than the previous one, or slightly worse to escape from local optima.

    It uses a :ref:`cooling` which partially drives the acceptance probability. This is the probability\
    to accept a worse solution according to the temperature, the best solution found so far and the actual solution.

    Attributes
    ----------

    loss_func : Loss
        :ref:`lf` to optimize. must be of type f(x)=y

    search_space : Searchspace
        :ref:`sp` object containing bounds of the search space.

    f_calls : int
        Maximum number of :ref:`lf` calls

    cooling : Cooling
        :ref:`cooling` used to determine the probability of acceptance.

    max_iter : int
        Maximum iterations of the inner loop.
        Determines how long the algorithm should sample neighbors of a solution,\
        before decreasing the temperature.

    save : boolean, optional
        if True save results into a file

    verbose : boolean, default=True
        Algorithm verbosity

    See Also
    --------
    :ref:`meta` : Parent class defining what a Metaheuristic is
    :ref:`lf` : Describes what a loss function is in Zellij
    :ref:`sp` : Describes what a loss function is in Zellij


    Examples
    --------

    >>> from zellij.core.loss_func import Loss
    >>> from zellij.core.search_space import Searchspace
    >>> from zellij.strategies.simulated_annealing import Simulated_annealing
    >>> from zellij.strategies.utils.cooling import MulExponential
    >>> from zellij.utils.benchmark import himmelblau
    ...
    >>> labels = ["a","b","c"]
    >>> types = ["R","R","R"]
    >>> values = [[-5, 5],[-5, 5],[-5, 5]]
    >>> sp = Searchspace(labels,types,values)
    >>> lf = Loss()(himmelblau)
    ...
    >>> cooling = MulExponential(0.85,100,2,3)
    >>> sa = Simulated_annealing(lf, sp, 100, cooling, 1)
    ...
    >>> point = sp.random_point()[0]
    >>> sa.run(point, lf([point])[0])
    >>> sa.show()

    .. image:: ../_static/sa_sp_ex.png
        :width: 924px
        :align: center
        :height: 487px
        :alt: alternate text
    .. image:: ../_static/sa_res_ex.png
        :width: 924px
        :align: center
        :height: 487px
        :alt: alternate text
    """

    # Initialize simulated annealing
    def __init__(
        self, loss_func, search_space, f_calls, cooling, max_iter, verbose=True
    ):

        """__init__(self,loss_func, search_space, f_calls, cooling, max_iter, verbose=True)

        Initialize Genetic_algorithm class

        Parameters
        ----------
        loss_func : Loss
            Loss function to optimize. must be of type f(x)=y

        search_space : Searchspace
            Search space object containing bounds of the search space.

        f_calls : int
            Maximum number of loss_func calls

        cooling : Cooling
            Cooling schedule used to determine the probability of acceptance.

        max_iter : int
            Maximum iterations of the inner loop.
            Determines how long the algorithm should sample neighbors of a solution,\
            before decreasing the temperature.

        save : boolean, optional
            if True save results into a file

        verbose : boolean, default=True
            Algorithm verbosity


        """

        super().__init__(loss_func, search_space, f_calls, verbose)

        # Max iteration after each temperature decrease
        self.max_iter = max_iter

        # Cooling schedule
        self.cooling = cooling

        self.n_scores = []
        self.n_best = []

        self.record_temp = [self.cooling.cool()]
        self.record_proba = [0]

        self.file_created = False

    # RUN SA
    def run(self, X0=None, Y0=None, H=None, n_process=1):

        """run(X0=None, Y0=None, H=None, n_process=1)

        Runs SA

        Parameters
        ----------
        X0 : list[float], optional
            Initial solution. If None, a Fractal must be given (H!=None)
        Y0 : {int, float}, optional
            Score of the initial solution
            Determine the starting point of the chaotic map.
        H : Fractal, optional
            When used by FDA, a fractal corresponding to the current subspace is given
        n_process : int, default=1
            Determine the number of best solution found to return.

        Returns
        -------
        best_sol : list[float]
            Returns a list of the <n_process> best found points to the continuous format

        best_scores : list[float]
            Returns a list of the <n_process> best found scores associated to best_sol

        """

        self.loss_func.file_created = False

        if X0:
            self.X0 = X0
        elif H:
            self.X0 = H.center
        else:
            raise ValueError("No starting point given to Simulated Annealing")

        if Y0:
            self.Y0 = Y0
        else:
            logger.info("Simulated annealing evaluating initial solution")
            self.Y0 = self.loss_func(
                self.search_space.convert_to_continuous([self.X0], True)
            )[0]

        self.n_best.append(X0)
        self.n_scores.append(Y0)

        logger.info("Starting")
        logger.debug(f"Starting solution: {X0}, {Y0}")

        # Determine the number of iteration according to the function parameters
        logger.debug("Determining number of iterations")
        nb_iteration = self.cooling.iterations() * self.max_iter
        logger.info(f"Number of iterations: {nb_iteration}")

        self.build_bar(nb_iteration)

        # Initialize variable for simulated annealing
        # Best solution so far
        X = self.X0[:]

        # Best solution in the neighborhood
        X_p = X[:]

        # Current solution
        Y = X[:]

        # Initialize score
        cout_X = self.Y0
        cout_X_p = self.Y0

        T_actu = self.cooling.Tcurrent

        # Simulated annealing starting
        while T_actu and self.loss_func.calls < self.f_calls:
            iteration = 0
            while (
                iteration < self.max_iter
                and self.loss_func.calls < self.f_calls
            ):

                neighbors = self.search_space.get_neighbor(X, size=n_process)

                # Update progress bar
                self.pending_pb(len(neighbors))

                loss_values = self.loss_func(
                    neighbors,
                    temperature=self.record_temp[-1],
                    probability=self.record_proba[-1],
                )

                # Update progress bar
                self.update_main_pb(
                    len(neighbors), explor=False, best=self.loss_func.new_best
                )

                index_min = np.argmin(loss_values)
                Y = neighbors[index_min][:]
                cout_Y = loss_values[index_min]

                # Compute previous cost minus new cost
                delta = cout_Y - cout_X

                logger.debug(f"New model score: {cout_Y}")
                logger.debug(f"Old model score: {cout_X}")
                logger.debug(f"Best model score: {cout_X_p}")

                # If a better model is found do...
                if delta < 0:
                    X = Y[:]
                    cout_X = cout_Y
                    if cout_Y < cout_X_p:

                        # Print if best model is found
                        logger.debug("Best model found: YES ")

                        X_p = X[:]
                        cout_X_p = cout_X

                    else:
                        logger.debug("Best model found: NO ")

                    self.record_proba.append(0)

                else:
                    logger.debug("Best model found: NO ")

                    p = np.random.uniform(0, 1)
                    emdst = np.exp(-delta / T_actu)

                    self.record_proba.append(emdst)

                    logger.debug(f"Escaping :  p<exp(-df/T) -->{p} < {emdst}")

                    if p <= emdst:
                        X = Y[:]
                        cout_X = cout_Y
                    else:
                        Y = X[:]

                iteration += 1
                self.meta_pb.update()

                logger.debug(
                    f"ITERATION: {self.loss_func.calls}/{self.f_calls}"
                )

                self.record_temp.append(T_actu)

                # Save file
                if self.loss_func.save:
                    if not self.file_created:
                        self.sa_save = os.path.join(
                            self.loss_func.outputs_path, "sa_best.csv"
                        )
                        with open(self.sa_save, "w") as f:
                            f.write(
                                ",".join(e for e in self.search_space.labels)
                                + ",loss,temperature,probability\n"
                            )
                            f.write(
                                ",".join(str(e) for e in self.X0)
                                + ","
                                + str(self.Y0)
                                + ","
                                + str(self.cooling.T0)
                                + ",0\n"
                            )
                            self.file_created = True

                    with open(self.sa_save, "a") as f:
                        f.write(
                            ",".join(str(e) for e in X)
                            + ","
                            + str(cout_X)
                            + ","
                            + str(self.record_temp[-1])
                            + ","
                            + str(self.record_proba[-1])
                            + "\n"
                        )

                self.n_scores.append(cout_X)
                self.n_best.append(X)

            T_actu = self.cooling.cool()

        # print the best solution from the simulated annealing
        logger.info(f"Best parameters: {X_p} score: {cout_X_p}")
        logger.info("Ending")

        best_idx = np.argpartition(self.loss_func.all_scores, n_process)
        best = [self.loss_func.all_solutions[i] for i in best_idx[:n_process]]
        min = [self.loss_func.all_scores[i] for i in best_idx[:n_process]]

        self.cooling.reset()
        self.close_bar()
        return best, min

    def show(self, filepath="", save=False):

        """show(self, filename=None)

        Plots solutions and scores computed during the optimization

        Parameters
        ----------
        filepath : str, default=""
            If a filepath is given, the method will read files insidethe folder and will try to plot contents.

        save : boolean, default=False
            Save figures
        """

        data_all, all_scores = super().show(filepath, save)

        if filepath:

            path_sa = os.path.join(filepath, "outputs", "sa_best.csv")
            data_sa = pd.read_table(path_sa, sep=",", decimal=".")
            sa_scores = data_sa["loss"].to_numpy()

            temperature = data_sa["temperature"].to_numpy()
            probability = data_sa["probability"].to_numpy()

        else:
            data_sa = self.n_best
            sa_scores = np.array(self.n_scores)

            temperature = np.array(self.record_temp)
            probability = np.array(self.record_proba)

        argmin = np.argmin(sa_scores)

        f, (l1, l2) = plt.subplots(2, 2, figsize=(19.2, 14.4))

        ax1, ax2 = l1
        ax3, ax4 = l2

        ax1.plot(list(range(len(sa_scores))), sa_scores, "-")
        argmin = np.argmin(sa_scores)
        all_argmin = np.argmin(all_scores)

        ax1.scatter(
            argmin,
            sa_scores[argmin],
            color="red",
            label="Best score: " + str(sa_scores[argmin]),
        )
        ax1.scatter(
            0,
            sa_scores[0],
            color="green",
            label="Initial score: " + str(sa_scores[0]),
        )

        ax1.set_title("Simulated annealing")
        ax1.set_xlabel("Iteration")
        ax1.set_ylabel("Score")
        ax1.legend(loc="upper right")

        if len(all_scores) < 100:
            s = 5
        else:
            s = 2500 / len(all_scores)

        ax2.scatter(list(range(len(all_scores))), all_scores, s=s)
        ax2.scatter(
            all_argmin,
            all_scores[all_argmin],
            color="red",
            label="Best score: " + str(sa_scores[argmin]),
        )
        ax2.scatter(
            0,
            sa_scores[0],
            color="green",
            label="Initial score: " + str(sa_scores[0]),
        )

        ax2.set_title("All evaluated solutions")
        ax2.set_xlabel("Solution ID")
        ax2.set_ylabel("Score")
        ax2.legend(loc="upper right")

        ax3.plot(list(range(len(sa_scores))), temperature, "-")
        argmin = np.argmin(sa_scores)
        ax3.scatter(
            argmin,
            temperature[argmin],
            color="red",
            label="Best score: " + str(temperature[argmin]),
        )

        ax3.set_title("Temperature decrease")
        ax3.set_xlabel("Iteration")
        ax3.set_ylabel("Temperature")
        ax3.legend(loc="upper right")

        if len(sa_scores) < 100:
            s = 5
        else:
            s = 2500 / len(all_scores)

        ax4.scatter(list(range(len(sa_scores))), probability, s=s)
        argmin = np.argmin(sa_scores)
        ax4.scatter(
            argmin,
            probability[argmin],
            color="red",
            label="Best score: " + str(probability[argmin]),
        )

        ax4.set_title("Escaping probability")
        ax4.set_xlabel("Iteration")
        ax4.set_ylabel("Probability")
        ax4.legend(loc="upper right")

        if save:
            save_path = os.path.join(
                self.loss_func.plots_path, f"sa_summary.png"
            )

            plt.savefig(save_path, bbox_inches="tight")
            plt.close()
        else:
            plt.show()



# @Author: Thomas Firmin <ThomasFirmin>
# @Date:   2022-05-03T15:41:48+02:00
# @Email:  thomas.firmin@univ-lille.fr
# @Project: Zellij
# @Last modified by:   ThomasFirmin
# @Last modified time: 2022-05-03T16:00:13+02:00
# @License: CeCILL-C (http://www.cecill.info/index.fr.html)
# @Copyright: Copyright (C) 2022 Thomas Firmin


import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
import pandas as pd
import os

import logging

logger = logging.getLogger("zellij.space")

# To modify
def parallel_coordinates(
    frame,
    class_column,
    cols=None,
    ax=None,
    color=None,
    use_columns=False,
    xticks=None,
    colormap=None,
    **kwds,
):

    n = len(frame)
    class_col = frame[class_column]
    class_min = np.amin(class_col)
    class_max = np.amax(class_col)

    if cols is None:
        df = frame.drop(class_column, axis=1)
    else:
        df = frame[cols]

    ncols = len(df.columns)

    # determine values to use for xticks
    if use_columns is True:
        if not np.all(np.isreal(list(df.columns))):
            raise ValueError("Columns must be numeric to be used as xticks")
        x = df.columns
    elif xticks is not None:
        if not np.all(np.isreal(xticks)):
            raise ValueError("xticks specified must be numeric")
        elif len(xticks) != ncols:
            raise ValueError("Length of xticks must match number of columns")
        x = xticks
    else:
        x = range(ncols)

    fig = plt.figure()
    ax = plt.gca()

    Colorm = plt.get_cmap(colormap)

    for i in range(n):
        y = df.iloc[i].values
        kls = class_col.iat[i]
        ax.plot(
            x,
            y,
            color=Colorm((kls - class_min) / (class_max - class_min)),
            alpha=1 - (kls - class_min) / (class_max - class_min),
            **kwds,
        )

    for i in x:
        ax.axvline(i, linewidth=1, color="black")

    ax.set_xticks(x)
    ax.set_xticklabels(df.columns)
    ax.set_xlim(x[0], x[-1])
    ax.grid()

    bounds = np.linspace(class_min, class_max, 10)
    cax, _ = mpl.colorbar.make_axes(ax)
    cb = mpl.colorbar.ColorbarBase(
        cax,
        cmap=Colorm,
        spacing="proportional",
        ticks=bounds,
        boundaries=bounds,
        format="%.2f",
    )

    return fig


# TYPES: REAL: R, DISCRETE: D, CATEGORICAL: C, CONSTANT: K
class Searchspace:

    """Searchspace

    Searchspace is an essential class for Zellij. Define your search space with this object.

    Attributes
    ----------

    labels : list[str]
        Labels associated to each dimension of the search space
    types : list[{'R','D','C','K'}]
        Types associated to each dimension of the search space.

        * R : Real dimension (e.g. float)
        * D : Discrete dimension (e.g. int)
        * C : Categorical dimension (e.g. {'dog', 'cats', 'rabbit'})
        * K : Constant dimension, used to define an unvariable dimension.

         Used in space decomposition when a dimension cannot be decomposed because it became too small (e.g. discrete).\
         For a constant value, implementing it directly to the loss function is preferable.

    values : list[{[float, float], [int, int], [{str, int, float}, {str, int, float}...]}]
        Bounds associated to each dimension.

        * For R, D types, values must be of form [a:{float,int}, b:{float,int}] with a < b.
        * For C type, values must be of form [{str, int, float}, {str, int, float}...]
        * For K type, values can be of any type because it will not be used and only passed to the loss function.

    neighborhood : list[{float, int, -1}]
        Neighborhood  of a solution for each dimension. For a solution :math:`x=[v_1, v_2]`, bounds of the neighborhood of :math:`v_i` of :math:`x` are computed by :math:`[max(v_i-n, lower bound), min(v_i+n, upper bound)]`,\
        except for C dimension, a neighbor is computed by drawing a random value assoiated to the dimension. For K dimension there is no neighborhood.

        * For R type, the neighborhood can be a float, or an int.
        * For D type, the neighborhood is an int.
        * For C,K types, the neighborhood must be -1.


    sub_values : list[{[float, float], [int, int], [{str, int, float}, {str, int, float}...]}]
        When building a subspace. Original values are saved in sub_values, so it can easily convert a subspace into a mixed or continuous search space. See values
    n_variables : list
        Number of dimensions
    k_index : list[int]
        Indexes of constant dimensions.

    Methods
    -------
    random_attribute(self,size=1,replace=True, exclude=None)
        Draw random features from the search space. Features of type K, cannot be drawn (i.e. their probability are set to 0).

    random_value(self, attribute, size=1, replace = True, exclude=None)
        Draw random values of an attribute from the search space, using uniform distribution. Features of type K return their constant value.

    get_neighbor(self, point, size=1, attribute=None)
        Draw a neighbor of an initial solution, according to the search space bounds and dimensions types.

    random_point(self,size=1)
        Return a random point from the search space

    convert_to_continuous(self,points,reverse=False,sub_values=False)
        Convert given points from mixed to continuous, or, from continuous to mixed.

    general_convert(self)
        Convert the search space by building a continuous one.

    subspace(self,lo_bounds,up_bounds)
        Build a sub space according to the actual Searchspace using two vectors containing lower and upper bounds of the subspace.

    show(self,X,Y)
        Show solutions X associated to their values Y, according to the Searchspace

    _create_neighborhood(self,neighborhood)
        Create the neighborhood. See Searchspace.__init__

    _get_real_neighbor(self, x, i)
            Draw a neighbor of a Real attribute from the search space, using uniform distribution. According to its lower and upper bounds

    _get_discrete_neighbor(self, x, i)
            Draw a neighbor of a Discrete attribute from the search space, using discrete uniform distribution. According to its lower and upper bounds

    _get_categorical_neighbor(self, x, i)
            Draw a neighbor of a Categorical attribute from the search space, using discrete uniform distribution. According to all its possible value

    See Also
    --------
    LossFunc : Parent class for a loss function.

    Examples
    --------
    >>> from zellij.core.search_space import Searchspace
    >>> labels = ["learning rate","neurons","activation"]
    >>> types = ["R","D","C"]
    >>> values = [[-5.0, 5.0],[0, 20],["relu","tanh","sigmoid"]]
    >>> neighborhood = [0.5,3,-1]
    >>> sp = Searchspace(labels,types,values, neighborhood)

    """

    # Initialize the search space
    def __init__(self, labels, types, values, neighborhood=0.10):

        """__init__(self, labels, types, values, neighborhood = 0.10 )

        Parameters
        ----------
        label : list[str]
            Labels associated to each dimension of the search space
        type : list[{'R','D','C','K'}]
            Types associated to each dimension of the search space.
            R : Real dimension (e.g. float)
            D : Discrete dimension (e.g. int)
            C : Categorical dimension (e.g. {'dog', 'cats', 'rabbit'})
            K : Constant dimension, used to define an unvariable dimension. \
             Used in space decomposition when a dimension cannot be decomposed because it became too small (e.g. discrete).\
             For a constant value, implementing it directly to the loss function is preferable.

        values : list[{[float, float], [int, int], [{str, int, float}, {str, int, float}...]}]
            Bounds associated to each dimension.\
            For R, D types, values must be of form [a:{float,int}, b:{float,int}] with a < b.\
            For C type, values must be of form [{str, int, float}, {str, int, float}...]
            For K type, values can be of any type because it will not be used and only passed to the loss function.

        neighborhood : {float, list[{float, int, -1}]}, default=0.10
            Neighborhood  of a solution for each dimension.\
            For a solution x=[v1, v2], bounds of the neighborhood of vi of x are computed by [max(vi-n, lower bound), min(vi+n, upper bound)],\
            except for C dimension, a neighbor is computed by drawing a random value assoiated to the dimension. For K dimension there is no neighborhood.

            If a float, N, between 0 and 1 is given. The neighborhood will be computed according to precedent rules, n = N*(upper bounds - lower bounds).
            The neighborhood for R, D dimension will be a percentage of the dimension size.

            You can manually determine the neighborhood according to following rules:
                - neighborhood must be of type list[{float, int, -1}], each value of the list determine the type of neighborhood for each dimension:
                    - For R type, the neighborhood can be a float, or an int.
                    - For D type, the neighborhood is an int.
                    - For C,K types, the neighborhood must be -1.
        """

        ##############
        # ASSERTIONS #
        ##############

        valid_types = ["R", "D", "C", "K"]

        assert len(labels) > 0, "Labels length must be > 0"
        assert len(types) > 0, "Type length must be > 0"
        assert len(values) > 0, "Values length must be > 0"
        assert (
            len(labels) == len(types) == len(values)
        ), "Labels, types and values must have the same length"

        assert (isinstance(neighborhood, list) and len(neighborhood) > 0) or (
            isinstance(neighborhood, float) and 0 < neighborhood < 1
        ), "neighborhood must be of the form [float|int|-1, ...], float: type='R', int: type='D', -1: type='C' or 'K' or must be a 0 < float < 1"

        assert all(isinstance(l, str) for l in labels), "Labels must be strings"
        assert all(
            t in valid_types for t in types
        ), "Types must be equal to 'R', 'D', 'C' or 'K' "

        for v, t in zip(values, types):
            if t == "R":
                assert (
                    isinstance(v, list)
                    and len(v) == 2
                    and (isinstance(v[0], float) or isinstance(v[0], int))
                    and (isinstance(v[1], float) or isinstance(v[1], int))
                    and v[0] < v[1]
                ), f"Values of type 'R' must be of the form [a : int, b : int] or [a : float, b : float] and b > a, got {v}"

            elif t == "D":
                assert (
                    isinstance(v, list)
                    and len(v) == 2
                    and isinstance(v[0], int)
                    and isinstance(v[1], int)
                    and v[0] < v[1]
                ), f"Values of type 'D' must be of the form [a : int, b : int] and b > a, got {v}"

            elif t == "C":
                assert (
                    isinstance(v, list) and len(v) >= 2
                ), f"Values of type 'C' must be of the form [value 1, value 2,...], got {v}"

        if isinstance(neighborhood, list):
            for n, t in zip(neighborhood, types):
                if t == "R":
                    assert isinstance(n, float) or isinstance(
                        n, int
                    ), f"neighborhood of type 'R' must be a float or an int, got {n}"

                elif t == "D":
                    assert isinstance(
                        n, int
                    ), f"neighborhood of type 'R' must be an int, got {n}"

                else:
                    assert (
                        n == -1
                    ), f"neighborhood of type 'C' or 'K' must be equal to -1, got {n}"

        ##############
        # PARAMETERS #
        ##############

        self.labels = labels
        self.types = types
        self.values = values
        self.sub_values = None

        #############
        # VARIABLES #
        #############

        self.n_variables = len(labels)

        self.k_index = [i for i, x in enumerate(self.types) if x == "K"]

        # if list is given do nothing, if a percentage is given, compute the neighborhood of a solution is located in an area corresponding to x% of the search space
        self._create_neighborhood(neighborhood)

    # Create the neighborhood for each variables
    def _create_neighborhood(self, neighborhood):

        """_create_neighborhood(neighborhood)

        Create the neighborhood. See Searchspace.__init__

        Parameters
        ----------
        neighborhood : {float, list[{float, int, -1}]}

        """

        self.neighborhood = []

        if type(neighborhood) == float:

            for i in range(self.n_variables):
                if self.types[i] == "R":
                    self.neighborhood.append(
                        (self.values[i][1] - self.values[i][0]) * neighborhood
                    )
                elif self.types[i] == "D":
                    self.neighborhood.append(
                        int(
                            np.ceil(
                                (self.values[i][1] - self.values[i][0])
                                * neighborhood
                            )
                        )
                    )
                else:
                    self.neighborhood.append(-1)
        else:
            self.neighborhood = neighborhood

    # Return 1 or size=n random attribute from the search space, can exclude one attribute
    def random_attribute(self, size=1, replace=True, exclude=None):

        """random_attribute(size=1,replace=True, exclude=None)

        Draw random features from the search space. Features of type K, cannot be drawn (i.e. their probability are set to 0).

        Parameters
        ----------
        size : int, default=1
            Select randomly <size> features.
        replace : boolean, default=True
            Select randomly <size> features with replacement if True, without else.
            See numpy.random.choice
        exclude : str, default=None
            Exclude dimension of label==<exclude>.

        Returns
        -------

        _ : numpy.array(dtype=int)
            Array of selected dimension index.

        Examples
        --------
        >>> rand_att = sp.random_attribute(5)
        >>> print(f"Random Attributes: {rand_att}")
        Random Attributes: ['learning rate' 'activation' 'neurons' 'activation' 'activation']
        """

        try:
            index = self.labels.index(exclude)
            p = np.full(
                self.n_variables,
                1 / (self.n_variables - (len(self.k_index) + 1)),
            )
            p[index] = 0

        except ValueError:
            p = np.full(
                self.n_variables, 1 / (self.n_variables - len(self.k_index))
            )

        for i in self.k_index:
            p[i] = 0

        return np.random.choice(self.labels, size=size, replace=replace, p=p)

    # Return a or size=n random value from an attribute, can exclude one value
    def random_value(self, attribute, size=1, replace=True, exclude=None):

        """random_value(attribute, size=1, replace = True, exclude=None)

        Draw random values of an attribute from the search space, using uniform distribution. Features of type K return their constant value.

        Parameters
        ----------
        attribute : str
            Select dimension of label==<exclude> from which to draw a random value.
        size : int, default=1
            Select randomly <size> features.
        replace : boolean, default=True
            Select randomly <size> features with replacement if True, without else.
            See numpy.random.choice
        exclude : str, default=None
            Exclude dimension of label==<exclude>.

        Returns
        -------

        _ : numpy.array(dtype={int, float})
            Array of floats for R dimension, ints for D, indexes (int) for C.

        Examples
        --------

        Real values

        >>> rand_val = sp.random_value('learning rate',5)
        >>> print(f"Random 'learning rate': {rand_val}")
        Random 'learning rate': [-2.137716545134689, 2.088492759876778, 3.970466658083497, -3.665643582672203, 3.116564115026975]

        Discrete values

        >>> rand_val = sp.random_value('neurons',5)
        >>> print(f"Random 'neurons': {rand_val}")
        Random 'neurons': [3, 8, 15, 8, 5]

        Categorical values

        >>> rand_val = sp.random_value('activation',5)
        >>> print(f"Random 'activation': {rand_val}")
        Random 'activation': ['tanh' 'sigmoid' 'tanh' 'sigmoid' 'sigmoid']

        """

        index = self.labels.index(attribute)

        if self.types[index] == "R":
            return np.random.uniform(
                self.values[index][0], self.values[index][1], size=size
            ).tolist()

        elif self.types[index] == "D":
            return np.random.randint(
                self.values[index][0], self.values[index][1] + 1, size=size
            ).tolist()

        elif self.types[index] == "C":

            try:
                idx = self.values[index].index(exclude)
                p = np.full(
                    len(self.values[index]), 1 / (len(self.values[index]) - 1)
                )
                p[idx] = 0
            except ValueError:
                p = np.full(
                    len(self.values[index]), 1 / len(self.values[index])
                )

            return np.random.choice(
                self.values[index], size=size, replace=replace, p=p
            )

        else:
            return [self.values[index] for _ in range(size)]

    def _get_real_neighbor(self, x, i):

        """_get_real_neighbor(x, i)

        Draw a neighbor of a Real attribute from the search space, using uniform distribution. According to its lower and upper bounds

        Parameters
        ----------

        x : float
            Initial value
        i : int
            Dimension index

        Returns
        -------

        v : float
            Random neighbor of x

        """

        upper = np.min([x + self.neighborhood[i], self.values[i][1]])
        lower = np.max([x - self.neighborhood[i], self.values[i][0]])
        v = np.random.uniform(lower, upper)

        while v == x:
            v = np.random.uniform(lower, upper)

        return float(v)

    def _get_discrete_neighbor(self, x, i):

        """_get_discrete_neighbor(x, i)

        Draw a neighbor of a Discrete attribute from the search space, using discrete uniform distribution. According to its lower and upper bounds

        Parameters
        ----------

        x : float
            Initial value
        i : int
            Dimension index

        Returns
        -------

        v : int
            Random neighbor of x

        """

        upper = int(np.min([x + self.neighborhood[i], self.values[i][1]])) + 1
        lower = int(np.max([x - self.neighborhood[i], self.values[i][0]]))
        v = np.random.randint(lower, upper)

        while v == x:
            v = np.random.randint(lower, upper)

        return int(v)

    def _get_categorical_neighbor(self, x, i):

        """_get_categorical_neighbor(self, x, i)

        Draw a neighbor of a Categorical attribute from the search space, using discrete uniform distribution. According to all its possible value

        Parameters
        ----------

        x : float
            Initial value
        i : int
            Dimension index

        Returns
        -------

        v : float
            Random neighbor of x

        """

        idx = np.random.randint(len(self.values[i]))

        while self.values[i][idx] == x:
            idx = np.random.randint(len(self.values[i]))

        v = self.values[i][idx]

        return v

    # Return a neighbor of a given point in the search space, can select neighbor of a particular attribute
    def get_neighbor(self, point, size=1, attribute=None):

        """get_neighbor(point, size=1, attribute=None)

        Draw a neighbor of an initial solution, according to the search space bounds and dimensions types.

        Parameters
        ----------

        point : list[{int, float, str}, {int, float, str}...]
            Initial point.
        size : int, default=1
            Draw <size> neighbors of <point>.
        size : str, default=None
            Draw a neighbor of <point> at dimension of label <attribute>

        Returns
        -------

        points : list[list[{int, float, str}, {int, float, str}...], ...]
            List of neighbors of <point>.

        """

        points = []

        if attribute == None:
            for _ in range(size):
                attribute = self.random_attribute()
                index = self.labels.index(attribute)
                neighbor = point[:]

                if self.types[index] == "R":
                    neighbor[index] = self._get_real_neighbor(
                        point[index], index
                    )

                elif self.types[index] == "D":

                    neighbor[index] = self._get_discrete_neighbor(
                        point[index], index
                    )

                elif self.types[index] == "C":

                    neighbor[index] = self._get_categorical_neighbor(
                        point[index], index
                    )

                points.append(neighbor[:])
        else:

            index = self.labels.index(attribute)

            if self.types[index] == "R":
                for _ in range(size):
                    points.append(self._get_real_neighbor(point[index], index))

            elif self.types[index] == "D":
                for _ in range(size):
                    points.append(
                        self._get_discrete_neighbor(point[index], index)
                    )

            elif self.types[index] == "C":
                for _ in range(size):
                    points.append(
                        self._get_categorical_neighbor(point[index], index)
                    )

        return points

    # Return a random point of the search space
    def random_point(self, size=1):

        """random_point(size=1)

        Return a random point from the search space

        Parameters
        ----------

        size : int, default=1
            Draw <size> points.

        Returns
        -------

        points : list[list[{int, float, str}]]
            List of <point>.

        Examples
        --------

        >>> rand_pts = sp.random_point(3)
        >>> print(f"Random points: {rand_pts}")
        Random points: [[-3.830114043118622, 9, 'sigmoid'],
        ...             [3.065902630698311, 3, 'sigmoid'],
        ...             [-0.6839762230289024, 10, 'relu']]

        """

        points = []

        for i in range(size):
            new_point = []
            for l in self.labels:
                new_point.append(self.random_value(l)[0])
            points.append(new_point[:])

        return points

    # Convert a point to continuous, or convert a continuous point to a point from the search space
    def convert_to_continuous(self, points, reverse=False, sub_values=False):

        """convert_to_continuous(points,reverse=False,sub_values=False)

        Convert given points from mixed to continuous, or, from continuous to mixed.

        Parameters
        ----------

        points : {list[list[{int, float, str}, {int, float, str}...], ...], list[list[float, float...], ...]}
            List of points to convert
        reverse : boolean, default=False
            If False convert points from mixed to continuous, if True, from continuous to mixed
        sub_values : boolean, default=True
            If the search space is a subspace, uses the original values to convert if True, else uses its own bounds.
            See Searchspace

        Returns
        -------

        points : {list[list[{int, float, str}, {int, float, str}...], ...], list[list[float, float...], ...]}
            List of converted points. Points are list of float if converted to continuous, else list of mixed variables.

        """

        # Uses bounds from the original space if this object is a subspace.
        if sub_values and self.sub_values != None:
            val = self.sub_values

        # Use initial bounds to convert
        else:
            val = self.values

        res = []

        # Continuous to mixed
        if reverse:

            for point in points:
                converted = []
                for att in range(self.n_variables):

                    if self.types[att] == "R":

                        converted.append(
                            point[att] * (val[att][1] - val[att][0])
                            + val[att][0]
                        )

                    elif self.types[att] == "D":

                        converted.append(
                            int(
                                point[att] * (val[att][1] - val[att][0])
                                + val[att][0]
                            )
                        )

                    elif self.types[att] == "C":

                        n_values = len(val[att])
                        converted.append(val[att][int(point[att] * n_values)])

                    elif self.types[att] == "K":
                        converted.append(self.values[att])

                res.append(converted[:])

        # Mixed to continuous
        else:

            for point in points:
                converted = []
                for att in range(self.n_variables):

                    if self.types[att] == "R" or self.types[att] == "D":

                        converted.append(
                            (point[att] - val[att][0])
                            / (val[att][1] - val[att][0])
                        )

                    elif self.types[att] == "C":

                        idx = self.values[att].index(point[att])
                        n_values = len(val[att])

                        converted.append(idx / n_values)

                    elif self.types[att] == "K":
                        converted.append(1)

                res.append(converted[:])

        return res

    def general_convert(self):

        """general_convert()

        Convert the search space by building a continuous one.
        labels are identical, all types are converted to R, and all bounds are between [0,1].
        Build an adptated neighborhood

        Returns
        -------

        sp : Searchspace
            Continuous Searchspace.

        """

        label = self.labels[:]
        type = ["R"] * self.n_variables
        values = [[0, 1]] * self.n_variables
        neighborhood = []

        for att in range(self.n_variables):

            if self.types[att] == "R" or self.types[att] == "D":

                neighborhood.append(
                    self.neighborhood[att]
                    / (self.values[att][1] - self.values[att][0])
                )

            else:

                neighborhood.append(1)

        sp = Searchspace(label, type, values, neighborhood)

        return sp

    def subspace(self, lo_bounds, up_bounds):

        """convert_to_continuous(points,reverse=False,sub_values=False)

        Build a sub space according to the actual Searchspace using two vectors containing lower and upper bounds of the subspace.
        Transforms types to K if necessary.
        Builds an adaptated neighborhood to avoid large neighborhood compare to the subspace size.
        Categorical lower and upper bounds of the subspace are determined according to a slice of the vector containing values:

            * Original: ["dog", "cat", "rabbit", "horse"]

            * lo_bounds = ["dog"]

            * up_bounds = ["rabbit"]

            * Subspace: ["dog", "cat", "rabbit"]

        Parameters
        ----------

        lo_bounds : {list[list[{int, float, str}, {int, float, str}...], ...], list[list[float, float...], ...]}
            List of points to convert
        up_bounds : boolean, default=False
            If False convert points from mixed to continuous, if True, from continuous to mixed

        Returns
        -------

        subspace : Searchspace
            Return a subspace of the actual Searchspace.

        """

        new_values = []
        new_neighborhood = []
        new_types = self.types[:]

        for i in range(len(lo_bounds)):

            if lo_bounds[i] != up_bounds[i]:
                if self.types[i] == "R":
                    new_values.append(
                        [
                            float(np.max([lo_bounds[i], self.values[i][0]])),
                            float(np.min([up_bounds[i], self.values[i][1]])),
                        ]
                    )
                    new_neighborhood.append(
                        self.neighborhood[i]
                        * (new_values[-1][1] - new_values[-1][0])
                        / (self.values[i][1] - self.values[i][0])
                    )

                elif self.types[i] == "D":
                    new_values.append(
                        [
                            int(np.max([lo_bounds[i], self.values[i][0]])),
                            int(np.min([up_bounds[i], self.values[i][1]])),
                        ]
                    )
                    n_val = int(
                        self.neighborhood[i]
                        * (new_values[-1][1] - new_values[-1][0])
                        / (self.values[i][1] - self.values[i][0])
                    )
                    new_neighborhood.append(1 if n_val == 0 else n_val)

                elif self.types[i] == "C":
                    new_neighborhood.append(-1)

                    lo_idx = self.values[i].index(lo_bounds[i])
                    up_idx = self.values[i].index(up_bounds[i])

                    if lo_idx > up_idx:
                        inter = up_idx
                        up_idx = lo_idx
                        lo_idx = inter

                    new_values.append(self.values[i][lo_idx : up_idx + 1])

                elif self.types[i] == "K":
                    new_neighborhood.append(-1)
                    new_values.append(self.values[i])

            # If lower and upper bounds are equal create a constant
            else:
                if self.types[i] != "K":
                    new_values.append(lo_bounds[i])
                    new_types[i] = "K"
                    new_neighborhood.append(-1)
                else:
                    new_neighborhood.append(-1)
                    new_values.append(self.values[i])

        subspace = Searchspace(
            self.labels, new_types, new_values, new_neighborhood
        )

        if subspace.sub_values == None:
            subspace.sub_values = self.values
        else:
            subspace.sub_values = self.sub_values

        return subspace

    def show(self, X, Y, save=False, path=""):

        """show(X, Y, save=False, path="")

        Show solutions X associated to their values Y, according to the Searchspace

        Parameters
        ----------

        X : list[list[{int, float, str}, {int, float, str}...], ...]
            List of points to plot
        Y : list[{float, int}]
            Scores associated to X.

        Examples
        --------

        >>> import numpy as np
        >>> sp.show(sp.random_point(100), np.random.random(100))


        .. image:: ../_static/sp_show_ex.png
            :width: 924px
            :align: center
            :height: 487px
            :alt: alternate text
        """

        if not isinstance(X, pd.DataFrame):
            X = pd.DataFrame(X, columns=self.labels)

        argmin = np.argmin(Y)
        logger.info("Best individual")
        logger.info(X.iloc[argmin, :])
        logger.info(np.array(Y)[argmin])

        f, plots = plt.subplots(
            self.n_variables, self.n_variables, figsize=(19.2, 14.4)
        )
        f.suptitle("All evaluated solutions", fontsize=11)

        if len(X) < 100:
            s = 40
        else:
            s = 10000 / len(X)

        if self.types[0] == "C":
            X.iloc[:, 0].value_counts().plot(kind="bar", ax=plots[0, 0])
            plots[0, 0].set_yticks([])
            plots[0, 0].xaxis.tick_top()
            plots[0, 0].tick_params(axis="x", labelsize=7 / len(self.types[0]))
        else:
            plots[0, 0].hist(
                X.iloc[:, 0], 20, density=True, facecolor="g", alpha=0.75
            )
            plots[0, 0].set_yticks([])
            plots[0, 0].xaxis.tick_top()
            plots[0, 0].tick_params(axis="x", labelsize=7)
            plots[0, 0].set_xlim((self.values[0][0], self.values[0][1]))

        def onpick(event):
            ind = event.ind
            print("Selected point:\n", X.iloc[ind, :], Y[ind])

        for i in range(self.n_variables):

            if i > 0:

                if self.types[i] == "C":

                    sorter = self.values[i]
                    sorterIndex = dict(zip(sorter, range(len(sorter))))

                    new = (
                        X.iloc[:, i]
                        .value_counts()
                        .rename_axis("unique_values")
                        .reset_index(name="counts")
                    )
                    new["Rank"] = new["unique_values"].map(sorterIndex)
                    new.sort_values("Rank", inplace=True)
                    new.drop("Rank", 1, inplace=True)
                    new = new.set_index("unique_values")

                    new["counts"].plot.barh(ax=plots[i, i], facecolor="g")
                    plots[i, i].yaxis.tick_right()
                    plots[i, i].tick_params(
                        axis="y", labelsize=7 / len(self.types[i])
                    )
                    plots[i, i].set_ylabel("")

                else:
                    plots[i, i].hist(
                        X.iloc[:, i],
                        20,
                        density=True,
                        facecolor="g",
                        alpha=0.75,
                        orientation="horizontal",
                    )
                    plots[i, i].yaxis.tick_right()
                    plots[i, i].tick_params(axis="y", labelsize=7)
                    plots[i, i].set_ylim((self.values[i][0], self.values[i][1]))

            for j in range(i + 1, self.n_variables):

                plots[i, j].axis("off")

                if self.types[i] == "C" or self.types[j] == "C":

                    if self.types[i] == self.types[j]:
                        pass
                    else:
                        if self.types[i] == "C":
                            idx = i
                            idx2 = j
                            vert = True
                        else:
                            idx = j
                            idx2 = i
                            vert = False

                        data = []
                        for val in self.values[idx]:
                            data.append(
                                X.iloc[:, idx2].loc[X.iloc[:, idx] == val]
                            )

                        plots[j, i].boxplot(
                            data,
                            vert=vert,
                            flierprops=dict(
                                marker="o",
                                markerfacecolor="green",
                                markersize=0.1,
                                markeredgecolor="green",
                            ),
                            labels=self.values[idx],
                        )

                else:
                    try:
                        plots[j, i].tricontourf(
                            X.iloc[:, i], X.iloc[:, j], Y, 10, cmap="Greys_r"
                        )
                    except:
                        logger.warning("Triangularisation failed")
                    plots[j, i].scatter(
                        X.iloc[:, i],
                        X.iloc[:, j],
                        c=Y,
                        s=s,
                        alpha=0.4,
                        cmap="coolwarm_r",
                        picker=True,
                    )
                    plots[j, i].set_xlim((self.values[i][0], self.values[i][1]))
                    plots[j, i].set_ylim((self.values[j][0], self.values[j][1]))

                    plots[j, i].scatter(
                        X.iloc[argmin, i],
                        X.iloc[argmin, j],
                        c="cyan",
                        marker=(5, 2),
                        alpha=0.8,
                        s=150,
                    )

                if i == 0:
                    plots[j, i].set_ylabel(self.labels[j])

                if j == self.n_variables - 1:
                    plots[j, i].set_xlabel(self.labels[i])

                plots[j, i].set_xticks([])
                plots[j, i].set_yticks([])

        plt.subplots_adjust(
            left=0.050, bottom=0.050, right=0.970, top=0.970, wspace=0, hspace=0
        )

        if save:
            save_path = os.path.join(path, f"matrix_sp.png")

            plt.savefig(save_path, bbox_inches="tight")
            plt.close()
        else:
            f.canvas.mpl_connect("pick_event", onpick)
            plt.show()

        # for i in range(self.n_variables):
        #     if self.types[i] == "C":
        #         inter = X[self.labels[i]].astype("category").cat.codes
        #         X.drop(self.labels[i], axis=1)
        #         X[self.labels[i]] = (inter - inter.min()) / (inter.max() - inter.min())
        #     else:
        #         X[self.labels[i]] = (X[self.labels[i]] - self.values[i][0]) / (self.values[i][1] - self.values[i][0])
        #
        # dataf = X.iloc[:, : self.n_variables]
        # dataf["loss_value"] = Y
        # parallel_coordinates(dataf, "loss_value", colormap="viridis_r")
        # plt.show()



# @Author: Thomas Firmin <ThomasFirmin>
# @Date:   2022-05-03T15:41:48+02:00
# @Email:  thomas.firmin@univ-lille.fr
# @Project: Zellij
# @Last modified by:   ThomasFirmin
# @Last modified time: 2022-05-03T15:45:45+02:00
# @License: CeCILL-C (http://www.cecill.info/index.fr.html)
# @Copyright: Copyright (C) 2022 Thomas Firmin


import numpy as np
from zellij.core.metaheuristic import Metaheuristic
from zellij.core.fractals import Hypersphere

import logging

logger = logging.getLogger("zellij.PHS")

# Promising Hypersphere Search
class PHS(Metaheuristic):

    """PHS

    Promising Hypersphere Search  is an exploration algorithm comming from original FDA paper.
    It used to evaluate the center of an Hypersphere, and fixed points on each dimension arround this center.

    It works on a continuous searh space.

    Attributes
    ----------

    up_bounds : list
        List of float containing the upper bounds of the search space converted to continuous.
    lo_bounds : list
        List of float containing the lower bounds of the search space converted to continuous.
    center : float
        List of floats containing the coordinates of the search space center converted to continuous.
    radius : float
        List of floats containing the radius for each dimensions of the search space converted to continuous.

    Methods
    -------

    run(self, n_process=1)
        Runs PHS


    See Also
    --------
    Metaheuristic : Parent class defining what a Metaheuristic is.
    LossFunc : Describes what a loss function is in Zellij
    Searchspace : Describes what a loss function is in Zellij
    """

    def __init__(self, loss_func, search_space, f_calls, verbose=True):

        """__init__(self, loss_func, search_space, f_calls,verbose=True)

        Initialize PHS class

        Parameters
        ----------
        loss_func : Loss
            Loss function to optimize. must be of type f(x)=y

        search_space : Searchspace
            Search space object containing bounds of the search space.

        f_calls : int
            Maximum number of loss_func calls

        verbose : boolean, default=True
            Algorithm verbosity

        """

        super().__init__(loss_func, search_space, f_calls, verbose)

    def run(self, H=None, n_process=1):
        """run(H=None, n_process=1)

        Parameters
        ----------
        H : Fractal, default=None
            When used by FDA, a fractal corresponding to the current subspace is given
        n_process : int, default=1
            Determine the number of best solution found to return.

        Returns
        -------
        best_sol : list[float]
            Returns a list of the <n_process> best found points to the continuous format

        best_scores : list[float]
            Returns a list of the <n_process> best found scores associated to best_sol

        """

        if H:
            assert isinstance(H, Hypersphere), logger.error(
                f"PHS should use Hyperspheres, got {H.__class__.__name__}"
            )

        current_idx = len(self.loss_func.all_solutions)
        points = np.zeros((3, self.search_space.n_variables), dtype=float)

        self.build_bar(self.f_calls)

        # logging
        logger.info("Starting")

        current_idx = len(self.loss_func.all_solutions)
        points = np.empty((3, self.search_space.n_variables), dtype=float)

        # logging
        logger.info("Starting")

        self.pending_pb(3)

        if self.loss_func.calls < self.f_calls:

            radius_part = (
                H.inflation * H.radius / np.sqrt(self.search_space.n_variables)
            )

            points[0] = H.center
            points[1] = H.center + radius_part
            points[2] = H.center - radius_part

            points[points > 1] = 1
            points[points < 0] = 0

        logger.info(f"Evaluating points")
        scores = self.loss_func(
            self.search_space.convert_to_continuous(points, True, True)
        )

        self.update_main_pb(3, explor=True, best=self.loss_func.new_best)
        self.meta_pb.update(3)

        logger.info("Ending")

        scores = np.array(scores)
        idx = np.array(np.argsort(scores))[:n_process]

        # best solution found
        best_sol = points[idx]
        best_scores = scores[idx]

        self.close_bar()

        logger.info("CGS ending")

        return best_sol, best_scores

    def show(self, filepath="", save=False):

        """show(filename="")

        Plots solutions and scores evaluated during the optimization

        Parameters
        ----------
        filename : str, default=None
            If a filepath is given, the method will read the file and will try to plot contents.

        save : boolean, default=False
            Save figures
        """

        super().show(filepath, save)



# @Author: Thomas Firmin <ThomasFirmin>
# @Date:   2022-05-03T15:41:48+02:00
# @Email:  thomas.firmin@univ-lille.fr
# @Project: Zellij
# @Last modified by:   ThomasFirmin
# @Last modified time: 2022-05-03T15:44:46+02:00
# @License: CeCILL-C (http://www.cecill.info/index.fr.html)
# @Copyright: Copyright (C) 2022 Thomas Firmin


import numpy as np


def minimum(H, *arg, **kwargs):
    """best(H, *arg, **kwargs)

    Parameters
    ----------
    H : Fractal
        Fractal on which to compute the heuristic value.

    Returns
    -------
    _ : float
        Minimum score found inside the fractal
    """

    return H.min_score


def median(H, *args, **kwargs):
    """median(H, *args, **kwargs)

    Parameters
    ----------
    H : Fractal
        Fractal on which to compute the heuristic value.

    Returns
    -------
    _ : float
        Median of all scores computed inside the fractal
    """

    return np.median(H.all_scores)


def mean(H, *args, **kwargs):
    """mean(H,best_ind,best_sc)

    Parameters
    ----------
    H : Fractal
        Fractal on which to compute the heuristic value.

    Returns
    -------
    _ : float
        Mean score of all scores computed inside the fractal
    """

    return np.mean(H.all_scores)


def std(H, *args, **kwargs):

    """_std(H, *args, **kwargs)

    Parameters
    ----------
    H : Fractal
        Fractal on which to compute the heuristic value.

    Returns
    -------
    _ : float
        Standard deviation of all scores computed inside the fractal
    """

    return np.std(H.all_scores)


def dttcb(H, *args, **kwargs):

    """dttcb(H,best_ind,best_sc)

    Parameters
    ----------
    H : Fractal
        Fractal on which to compute the heuristic value.

    Returns
    -------
    _ : float
        Distance to the best found solution so far to the best found solution computed inside the fractal
    """

    best_ind = args[0]

    return np.min(
        np.array(H.all_scores)
        / np.linalg.norm(np.array(H.solutions) - np.array(best_ind), axis=1)
    )


def belief(H, gamma=0.5, *args, **kwargs):

    """belief(H,best_ind,best_sc)

    Parameters
    ----------
    H : Fractal
        Fractal on which to compute the heuristic value.

    gamma : float, default=0.5
        Influences the contribution of the father's best score and the mean of all scores inside the fractal.
        Higher gamma means higher contribution of the fractal's father.

    Returns
    -------
    _ : float
        Belief computed according to the influences of the father's score and the mean of all computed solutions inside the fractal
    """

    best_sc = args[1]

    if type(H.father.father) == str:
        H.father.score = 0

    ratio = np.array(H.all_scores) / best_sc
    return -(
        gamma * H.father.score
        + (1 - gamma) * np.mean(ratio * np.exp(1 - ratio))
    )
